"""
This module contains functions for calculating diagnostics for the meta-analysis. The diagnostics are saved to files
in the ``timewise_sup_data`` directory. The diagnostics are:

* ``statuses_per_chunk.json``: the status of the objects in each chunk
* ``catalog_matches_per_chunk.json``: the number of catalog matches in each chunk
* ``positional_outlier_mjds``: the MJDs of the positional outliers in each chunk

The functions are:

* :func:`calculate_positional_outlier_times` calculates the MJDs of the positional outliers in a given chunk
* :func:`get_statuses_per_chunk` gets the statuses of the objects in each chunk
* :func:`get_catalog_matches_per_chunk` gets the number of catalog matches in each chunk
* :func:`get_positional_outliers_times` gets the MJDs of the positional outliers in each chunk
* :func:`get_database_summary` gets a summary of the documents in the database

"""

import logging
import os
import json
from pathlib import Path
import pandas as pd
import tqdm
from datetime import datetime
from timewise.wise_data_base import WISEDataBase

from timewise_sup.environment import load_environment
from timewise_sup.mongo import DatabaseConnector
from timewise_sup.meta_analysis.catalog_match import get_catalog_match_mask


logger = logging.getLogger(__name__)


def get_statuses_per_chunk(
        base_name: str,
        database_name: str,
        wise_data: WISEDataBase
) -> dict:
    """
    Get the statuses of the objects in each chunk. The statuses are saved to a file in the ``timewise_sup_data``
    directory. If the file already exists, it is loaded from there. Otherwise, it is calculated and saved.

    :param base_name: base name of the WISE data
    :type base_name: str
    :param database_name: name of the database
    :type database_name: str
    :param wise_data: the WISE data
    :type wise_data: WISEDataBase
    :return: the statuses of the objects in each chunk
    :rtype: dict
    """
    logger.info(f"getting statuses per chunk for {base_name}")
    tsup_data_dir = load_environment("TIMEWISE_SUP_DATA")
    fn = os.path.join(tsup_data_dir, base_name, "statuses_per_chunk.json")

    if not os.path.isfile(fn):
        logger.debug(f"No file {fn}. Calculating")
        chunks = list(range(wise_data.n_chunks))

        logger.info("getting statusees")
        statusees = dict()
        for c in chunks:
            m = wise_data.chunk_map == c
            ids = wise_data.parent_sample.df.index[m]
            status = DatabaseConnector(base_name=base_name, database_name=database_name).get_status(ids)
            statusees[c] = list(status.status)

        logger.debug(f"saving under {fn}")
        with open(fn, "w") as f:
            json.dump(statusees, f)

    else:
        logger.debug(f"loading {fn}")
        with open(fn, "r") as f:
            statusees = json.load(f)

    return statusees


def get_catalog_matches_per_chunk(
        base_name: str,
        database_name: str,
        wise_data: WISEDataBase
) -> dict:
    """
    Get the number of catalog matches in each chunk. The number of catalog matches are saved to a file in the
    ``timewise_sup_data`` directory. If the file already exists, it is loaded from there. Otherwise, it is calculated
    and saved.

    :param base_name: base name of the WISE data
    :type base_name: str
    :param database_name: name of the database
    :type database_name: str
    :param wise_data: the WISE data
    :type wise_data: WISEDataBase
    :return: the number of catalog matches in each chunk
    :rtype: dict
    """
    logger.info("getting catalog matches per chunk")

    tsup_data_dir = load_environment("TIMEWISE_SUP_DATA")
    fn = os.path.join(tsup_data_dir, base_name, "catalog_matches_per_chunk.json")

    if not os.path.isfile(fn):
        logger.debug(f"No file {fn}. Calculating")

        chunks = list(range(wise_data.n_chunks))

        matches_per_chunk = dict()
        for c in chunks:
            m = wise_data.chunk_map == c
            ids = wise_data.parent_sample.df.index[m]
            chunk_match_mask = get_catalog_match_mask(base_name, database_name, ids)

            chunk_matches = dict()
            for catalogue_name in chunk_match_mask.columns:
                chunk_matches[catalogue_name] = int(chunk_match_mask[catalogue_name].sum())

            matches_per_chunk[c] = chunk_matches

        logger.debug(f"saving to {fn}")
        with open(fn, "w") as f:
            json.dump(matches_per_chunk, f)

    else:
        logger.debug(f"loading {fn}")
        with open(fn, "r") as f:
            matches_per_chunk = json.load(f)

    return matches_per_chunk


def calculate_positional_outlier_times(
        wise_data: WISEDataBase,
        chunk_number: int
) -> list:
    """
    Use :class:`timewise` to calculate the MJDs of the positional outliers in a given chunk.
    See the `documentation
    <https://timewise.readthedocs.io/en/latest/api.html#timewise.wise_data_base.WISEDataBase.get_position_mask>`_
    for more information.

    :param wise_data: the WISE data
    :type wise_data: WISEDataBase
    :param chunk_number: the chunk number
    :type chunk_number: int
    :return: the MJDs of the positional outliers in the chunk
    :rtype: list
    """
    logging.getLogger("timewise").setLevel(logging.getLogger("timewise_sup").getEffectiveLevel())
    unbinned_lcs = wise_data.get_unbinned_lightcurves(chunk_number=chunk_number)
    position_masks = wise_data.get_position_mask(service="tap", chunk_number=chunk_number)

    mjds = list()

    for ind, position_mask in tqdm.tqdm(position_masks.items(), desc="going through lightcurves"):
        lc = unbinned_lcs[unbinned_lcs[wise_data._tap_orig_id_key] == int(ind)]
        mjds.extend(list(lc.loc[position_mask].mjd.values))

    return mjds


def get_positional_outliers_times(
        base_name,
        wise_data: WISEDataBase
) -> dict:
    """
    Get the MJDs of the positional outliers in each chunk. The MJDs are saved to files in the ``timewise_sup_data``
    directory. If the files already exist, they are loaded from there. Otherwise, they are calculated and saved.

    :param base_name: base name of the WISE data
    :type base_name: str
    :param wise_data: the WISE data
    :type wise_data: WISEDataBase
    :return: the MJDs of the positional outliers in each chunk
    :rtype: dict
    """
    logger.info(f"getting positional outlier times")

    tsup_data = load_environment("TIMEWISE_SUP_DATA")
    cache_dir = os.path.join(tsup_data, base_name, "positional_outlier_mjds")
    os.makedirs(cache_dir, exist_ok=True)

    mjds_per_chunk = dict()

    for c in tqdm.tqdm(range(wise_data.n_chunks), desc="going through chunks"):
        fn = os.path.join(cache_dir, f"{c}.json")

        if not os.path.isfile(fn):
            logger.debug(f"file {fn} not found. Calculating")
            mjds = calculate_positional_outlier_times(wise_data, c)
            logger.debug(f"saving to {fn}")

            with open(fn, "w") as f:
                json.dump(mjds, f)

        else:
            logger.debug(f"loading {fn}")
            with open(fn, "r") as f:
                mjds = json.load(f)

        mjds_per_chunk[c] = mjds

    return mjds_per_chunk


def get_database_summary(base_name: str, database_name: str) -> pd.DataFrame:
    """
    Get a summary of the documents in the database. Count documents of the following units:
    - ```T2CatalogMatch```
    - ```T2DigestRedshifts```
    - ```T2BayesianBlocks```
    - ```T2DustEchoEval```
    Count their numbers for each document code.

    :param base_name: base name of the wise data
    :type base_name: str
    :param database_name: name of the database
    :type database_name: str
    :return: dataframe with the summary
    :rtype: pd.DataFrame
    """
    logger.info(f"getting database summary for {base_name}")
    t2col = DatabaseConnector(base_name=base_name, database_name=database_name).t2collection
    codes = t2col.distinct("code")
    units = ["T2CatalogMatch", "T2DigestRedshifts", "T2BayesianBlocks", "T2DustEchoEval"]
    summary = pd.DataFrame(index=codes, columns=units)
    unit_codes = [(unit, code) for unit in units for code in codes]
    for unit, code in tqdm.tqdm(unit_codes, desc="going through units and codes"):
        summary.loc[code, unit] = t2col.count_documents({"unit": unit, "code": code})
    t = datetime.now().strftime("%Y-%m-%d_%H-%M-%S")
    fn = (
            Path(load_environment("TIMEWISE_SUP_DATA")) /
            base_name /
            database_name /
            f"database_summary_{t}.csv"
    )
    fn.parent.mkdir(parents=True, exist_ok=True)
    logger.debug(f"saving to {fn}")
    summary.to_csv(fn)
    return summary



"""
Calculate the IR energy of the flares. This is done by integrating the luminosity over the duration of the flare.

The IR energy is calculated and stored in a json file. The file is named
ir_eng_status{status}.json, where status is the status of the flare. The file is stored in the directory
`timewise_sup_data/{base_name}`. If the file already exists, it is loaded from there. If it does not exist, it is
calculated and stored there.

* :func:`ir_energy_integral` calculates the IR energy for a given light curve
* :func:`calculate_ir_energy` calculates the IR energy for a given status
* :func:`get_ir_energy_status` calculates the IR energy for a given status
* :func:`get_ir_energy_index` calculates the IR energy for a given index

"""

import logging
import os
import json
from collections.abc import Mapping
import numpy as np
import pandas as pd
from scipy import integrate
from timewise.wise_data_base import WISEDataBase

from timewise_sup.environment import load_environment
from timewise_sup.mongo import DatabaseConnector, Index, Status
from timewise_sup.meta_analysis.luminosity import (
    get_ir_luminosities,
    luminosity_key,
    luminosity_err_key,
    ref_time_key,
    get_ref_time
)
from timewise_sup.meta_analysis.integrate_time import time_integral


logger = logging.getLogger(__name__)

ir_energy_key = "ir_energy_erg"
ir_energy_err_key = "ir_energy_erg_err"


def calculate_ir_energy(
        base_name: str,
        database_name: str,
        wise_data: WISEDataBase,
        status: Status,
        redshifts: pd.DataFrame | None = None,
) -> dict:
    """
    Calculate the IR energy of the flares. This is done by integrating the luminosity over the duration of the flare.

    :param base_name: base name for storage directories
    :type base_name: str
    :param database_name: name of the database
    :type database_name: str
    :param wise_data: instance of WISEDataBase
    :type wise_data: WISEDataBase
    :param status: status of the flare
    :type status: Status
    :param redshifts:
        redshift of the object, defaults to getting the redshift from the AMPEL catalog crossmatch,
        needs keys `z` and `z_unc`
    :type redshifts: pandas.DataFrame | None
    :return: dictionary with the index as key and the IR energy as value
    :rtype: dict
    """
    database_connector = DatabaseConnector(base_name=base_name, database_name=database_name)
    ids = database_connector.get_ids(status)
    indices = np.atleast_1d(ids)
    logger.info(f"calculating luminosities for {len(indices)} objects")
    luminosity_dict = get_ir_luminosities(base_name, database_name, wise_data, status, redshifts=redshifts)
    excess_mjd = database_connector.get_excess_mjd(indices)

    engs = {}  # type: dict[str, dict[str, float] | None]

    for k, lum_info in luminosity_dict.items():

        if lum_info is None:
            logger.warning(f"no luminosity information for {k}. Can not calculate IR energy.")
            engs[k] = None
            continue

        ref_time = lum_info["ref_time_mjd"]
        lc = pd.DataFrame(lum_info["lightcurve"])
        t_start = excess_mjd.loc[k, ["W1_excess_start_mjd", "W2_excess_start_mjd"]].min() - ref_time
        t_end = excess_mjd.loc[k, ["W1_excess_end_mjd", "W2_excess_end_mjd"]].max() - ref_time
        integral = time_integral(
            lc=lc,
            key=luminosity_key,
            err_key=luminosity_err_key,
            t_start=t_start,
            t_end=t_end,
            time_key=ref_time_key,
        )

        ieng = dict()
        for b in ["W1", "W2"]:
            ieng[f"{b}_{ir_energy_key}"] = integral[f"{b}_{luminosity_key}_integrated"]
            ieng[f"{b}_{ir_energy_err_key}"] = integral[f"{b}_{luminosity_key}_integrated_err"]
        ieng[ir_energy_key] = np.sum([ieng[f"{b}_{ir_energy_key}"] for b in ["W1", "W2"]])
        ieng[ir_energy_err_key] = np.sqrt(np.sum([ieng[f"{b}_{ir_energy_err_key}"] ** 2 for b in ["W1", "W2"]]))
        ieng["ir_energy_is_lower_limit"] = bool(np.any([excess_mjd.loc[k, f"{b}_flare_ended"] for b in ["W1", "W2"]]))
        ieng["z"] = lum_info["z"]
        ieng["z_err"] = lum_info["z_err"]
        engs[k] = ieng

    return engs


def get_ir_energy_status(
        base_name: str,
        database_name: str,
        status: Status,
        redshifts: pd.DataFrame | None = None,
        wise_data: WISEDataBase | None = None,
        force_new: bool = False,
) -> dict:
    """
    Get the IR energy of the flares per status. If the cache file `ir_eng_status{status}.json` exists, it is loaded from there.
    If it does not exist, it is calculated by :func:`calculate_ir_energy` and stored there.

    :param base_name: base name for storage directories
    :type base_name: str
    :param database_name: name of the database
    :type database_name: str
    :param status: status for which the IR energy should be calculated
    :type status: Status
    :param redshifts:
        redshift of the object, defaults to getting the redshift from the AMPEL catalog crossmatch,
        needs keys `z` and `z_unc`
    :type redshifts: pandas.DataFrame | None
    :param wise_data: instance of WISEDataBase
    :type wise_data: WISEDataBase
    :param force_new: if True, the IR energy is calculated and stored in the cache file, even if it already exists
    :type force_new: bool
    :return: dictionary with the index as key and the IR energy as value
    :rtype: dict
    """
    logger.info(f"getting bolometric luminosities for status {status}")
    tsup_data_dir = load_environment("TIMEWISE_SUP_DATA")
    fn = os.path.join(tsup_data_dir, base_name, f"ir_eng_status{status}.json")

    if (not os.path.isfile(fn)) or force_new:

        if wise_data is None:
            raise ValueError("wise_data must be given when calculating IR energies!")
        else:
            logger.debug(f"No file {fn}.")
            logger.info("Making bolometric luminosities")
            engs = calculate_ir_energy(
                base_name=base_name,
                database_name=database_name,
                wise_data=wise_data,
                status=status,
                redshifts=redshifts,
            )

        with open(fn, "w") as f:
            json.dump(engs, f)

    else:
        logger.debug(f"loading {fn}")
        with open(fn, "r") as f:
            engs = json.load(f)

        if redshifts is None:
            connector = DatabaseConnector(base_name=base_name, database_name=database_name)
            indices = connector.get_ids(status)
            redshifts = connector.get_redshift(indices)

        for i, eng_info in engs.items():
            if eng_info is None:
                logger.warning(f"missing luminosity for {i}")

            elif eng_info["z"] != redshifts.loc[i, "z"]:
                raise ValueError(f"redshift mismatch for {i}: {eng_info['z']} != {redshifts.loc[i, 'z']}")

    return engs


def get_ir_energy_index(
        base_name: str,
        database_name: str,
        index: Index,
        wise_data: WISEDataBase,
        forcenew: bool = False,
) -> dict:
    """
    Get the IR energy of the flares per index by getting the status and calling :func:`get_ir_energy_status`
    for each status.

    :param base_name:
    :type base_name: str
    :param database_name:
    :type database_name: str
    :param index: index of the lightcurve
    :type index: Index
    :param forcenew: if True, the IR energy is calculated and stored in the cache file, even if it already exists
    :type forcenew: bool
    :return: dictionary with the index as key and the IR energy as value
    :rtype: dict
    """

    statuses = DatabaseConnector(base_name=base_name, database_name=database_name).get_status(index)

    engs_all = dict()

    for status in statuses.status.unique():
        eng = get_ir_energy_status(base_name, database_name, status, wise_data, forcenew)
        selected_lums = {k: v for k, v in eng.items() if k in statuses.index[statuses.status == status]}
        engs_all.update(selected_lums)

    return engs_all



"""
Make histograms of the chi2 values for each chunk using the ``WISEDataDESYCluster.make_chi2_plot()`` method.

* :func:`make_timewise_chi2_histograms` makes histograms of the chi2 values for each chunk using the
``WISEDataDESYCluster.make_chi2_plot()`` method.
"""

import logging
import numpy as np
from timewise import WISEDataDESYCluster

from timewise_sup.meta_analysis.agn import get_agn_mask_from_color


logger = logging.getLogger(__name__)


def make_timewise_chi2_histograms(
        base_name: str,
        wise_data: WISEDataDESYCluster,
        chunks: list | None = None,
        plot_options: dict | None = None
):
    """
    Make histograms of the chi2 values for each chunk using the ``WISEDataDESYCluster.make_chi2_plot()`` method.
    See the ``timewise`` `documentation <https://timewise.readthedocs.io/en/latest/api.html#timewise.wise_bigdata_desy_cluster.WISEDataDESYCluster.make_chi2_plot>`_
    for more information.

    :param base_name: base name for storage directories
    :type base_name: str
    :param wise_data: instance of WISEDataDESYCluster
    :type wise_data: WISEDataDESYCluster
    :param chunks: list of chunks to use
    :type chunks: list
    :param plot_options: options for the plot, see ``WISEDataDESYCluster.make_chi2_plot()``
    :type plot_options: dict
    """

    logger.info("making timewise chi2 histograms")

    if chunks is None:
        chunks = list(range(wise_data.n_chunks))

    logger.debug(f"using chunks {chunks}")

    wise_mag_keys = ["W1_mag", "W2_mag", "W3_mag"]
    if np.any([k not in wise_data.parent_sample.default_keymap for k in wise_mag_keys]):
        logger.warning(f"One of {','.join(wise_mag_keys)} not in ParentSample's default keymap: \n"
                       f"{wise_data.parent_sample.default_keymap}.\n"
                       f"Not making separate histograms for AGNs.")
        index_mask = {}

    else:
        agn_mask, w1w2, w2w3 = get_agn_mask_from_color(base_name, wise_data)
        agn_indices = agn_mask.index[agn_mask].astype(str)
        non_agn_indices = agn_mask.index[~agn_mask].astype(str)
        index_mask = {
            "AGNs": agn_indices,
            "non-AGNs": non_agn_indices,
        }

    if plot_options is None:
        plot_options = {}

    wise_data.make_chi2_plot(
        chunks=chunks,
        load_from_bigdata_dir=True,
        lum_key="_flux_density",
        index_mask=index_mask,
        save=True,
        **plot_options
    )



"""
This module contains the :class:`TimewiseSUPConfig` and :class:`TimewiseSUPConfigLoader` classes which are used to
load the configuration for ``timewise-sup``.

* :class:`TimewiseSUPConfig` is a subclass of :class:`timewise.config_loader.TimewiseConfig` used to run the config
* :class:`TimewiseSUPConfigLoader` is a subclass of :class:`timewise.config_loader.TimewiseConfigLoader` used to load the config from a YAML file
"""

import logging
import inspect
from pydantic import validator
from timewise.config_loader import TimewiseConfig, TimewiseConfigLoader

from timewise_sup.ampel_conf import create_ampel_config_file
from timewise_sup.analyse_lightcurves.bayesian_blocks import bayesian_blocks
from timewise_sup.meta_analysis.baseline_subtraction import get_baseline_subtracted_lightcurves
from timewise_sup.plots.plot_lightcurves import plot_sample_lightcurves
from timewise_sup.analyse_lightcurves.cluster import run_bayesian_blocks_on_cluster
from timewise_sup.plots.diagnostic_plots import chunk_distribution_plots, positional_outliers


logger = logging.getLogger(__name__)


functions = {
    "create_ampel_config_file": create_ampel_config_file,
    "bayesian_blocks": bayesian_blocks,
    "baseline_subtraction": get_baseline_subtracted_lightcurves,
    "plot": plot_sample_lightcurves,
    "run_cluster": run_bayesian_blocks_on_cluster,
    "plot_distributions_among_chunks": chunk_distribution_plots,
    "plot_positional_outliers": positional_outliers
}

_functions_str = "\n".join(
    [f"\n\t* ``{fct_name}``: :func:`{fct.__module__}.{fct.__name__}` " for fct_name, fct in functions.items()]
)


class TimewiseSUPConfigLoader(TimewiseConfigLoader):
    __doc__ = f"""
    Loads the config for ``timewise-sup`` from a YAML file.
    The config is validated and parsed into a :class:`TimewiseSUPConfig` object.
    In addition to the parameters of :class:`timewise.config_loader.TimewiseConfigLoader`, the following parameters
    are available:
    
    :param timewise_sup_instructions: a list of dictionaries containing the instructions for ``timewise-sup``.
        Each dictionary must contain a single key-value pair, where the key is the name of the function to be called
        and the value is a dictionary of arguments for the function. The arguments must be valid for the function.
        The functions are called in the order they appear in the list. The possible functions are:
        {_functions_str}
    :type timewise_sup_instructions: list
    :param database_name: the name of the database in the ``MongoDB`` to be used for the analysis
    :type database_name: str
    """

    timewise_sup_instructions: list[dict] = list()
    database_name: str

    def parse_config(self):
        timewise_config = super().parse_config()
        return TimewiseSUPConfig(
            database_name=self.database_name,
            timewise_sup_instructions=self.timewise_sup_instructions,
            **timewise_config.dict()
        )


class TimewiseSUPConfig(TimewiseConfig):
    """
    Config for ``timewise-sup``.
    In addition to running the ``timewise`` config, it runs the functions specified in
    :attr:`timewise_sup_instructions` in the order they appear in the list.
    """

    timewise_sup_instructions: list[dict] = list()
    database_name: str

    @validator("timewise_sup_instructions")
    def validate_timewise_sup_instructions(cls, v: list[dict]):
        for instructions in v:
            for fct_name, arguments in instructions.items():

                if fct_name not in functions:
                    available = ", ".join(list(functions.keys()))
                    raise ValueError(f"timewise-sup has no function {fct_name}! Must be either of {available}")

                fct = functions[fct_name]
                # this is a mypy bug: https://github.com/python/mypy/issues/10740

                signature = inspect.signature(fct)  # type: ignore
                param_list = list(signature.parameters)
                # check if the function needs default arguments
                _arguments = arguments or dict()  # supply empty dict if arguments is None
                for k in ["base_name", "database_name", "wise_data"]:
                    if k in param_list:
                        # enter dummy string in arguments
                        _arguments[k] = ""
                # check validity of arguments
                try:
                    signature.bind(**_arguments)
                except TypeError as e:
                    raise ValueError(f"{fct_name}: {e}!")

        return v

    def run_config(self):
        logger.info("running config")
        super().run_config()

        for instructions in self.timewise_sup_instructions:
            for fct_name, arguments in instructions.items():
                _arguments = arguments or dict()  # supply empty dict if arguments is None
                fct = functions[fct_name]
                params = list(inspect.signature(fct).parameters)

                if "base_name" in params:
                    _arguments["base_name"] = self.wise_data.base_name
                if "database_name" in params:
                    _arguments["database_name"] = self.database_name
                if "wise_data" in params:
                    _arguments["wise_data"] = self.wise_data

                logger.info(f"running {fct_name} with arguments {_arguments}")
                fct(**_arguments)

        logger.info("successfully ran config")



"""
Functions to perform integration of a lightcurve over the duration of the flare

* :func:`time_integral` calculates the integral of a lightcurve over the duration of the flare
"""

import logging
import pandas as pd
from scipy import integrate
import numpy as np

from timewise_sup.mongo import DatabaseConnector, Index


logger = logging.getLogger(__name__)


def time_integral(
        lc: pd.DataFrame,
        key: str,
        err_key: str,
        t_start: float,
        t_end: float,
        time_key: str = "mean_mjd",
) -> dict[str, float]:
    mjd = lc[time_key]
    integral = {}

    for b in ["W1", "W2"]:
        mask = (mjd >= t_start) & (mjd <= t_end)
        fit_val = lc[f"{b}_{key}"][mask]
        fit_mjd = mjd[mask]

        # integrate using trapezoidal rule
        i_integral = integrate.trapezoid(fit_val, fit_mjd * 24 * 3600)
        integral[f"{b}_{key}_integrated"] = i_integral

        # try to estimate the error:
        # maximum of difference when calculating the integral with the upper and lower error
        fit_val_err = lc[f"{b}_{err_key}"][mask]
        fit_val_err_up = fit_val + fit_val_err
        fit_val_err_lo = fit_val - fit_val_err
        errors = np.array([integrate.trapezoid(e, fit_mjd * 24 * 3600) for e in [fit_val_err_up, fit_val_err_lo]])
        i_integral_err = max(np.abs(i_integral - errors))
        integral[f"{b}_{key}_integrated_err"] = i_integral_err

    return integral


def get_flare_time(
        base_name: str,
        database_name: str,
        index: Index,
) -> pd.DataFrame:
    """
    Get the flare time for a given index. The start and end times are the median of the start and end of the
    IR excess periods in W1 and W2. The lower limit is True if the flare di not end in W1 or W2.

    :param base_name: base name of the wise data
    :type base_name: str
    :param database_name: name of the database
    :type database_name: str
    :param index: index of the object
    :type index: int
    :return: flare time
    :rtype: dict
    """
    database_connector = DatabaseConnector(base_name=base_name, database_name=database_name)
    excess_times = database_connector.get_excess_mjd(index=index)
    start_times = excess_times[["W1_excess_start_mjd", "W2_excess_start_mjd"]].median(axis=1)
    end_times = excess_times[["W1_excess_end_mjd", "W2_excess_end_mjd"]].median(axis=1)
    lower_limit = ~excess_times[["W1_flare_ended", "W2_flare_ended"]].all(axis=1)
    return pd.DataFrame({
        "start_time": start_times,
        "end_time": end_times,
        "lower_limit": lower_limit,
    })



"""
Plot single lightcurves or all lightcurves of a given status.

* :func:`plot_sample_lightcurves` plots all lightcurves of a given status
* :func:`plot_single_lightcurve` plots a single lightcurve
* :func:`make_lightcurve_plot` is the actual plotting function
"""

import os.path

import matplotlib.pyplot as plt
import matplotlib as mpl
import logging
import pandas as pd
from timewise.wise_data_base import WISEDataBase

from timewise_sup.meta_analysis.flux import get_ir_flux, get_ir_flux_index, flux_key, flux_err_key
from timewise_sup.plots import plots_dir, bandcolors


logger = logging.getLogger(__name__)


def plot_sample_lightcurves(
        base_name: str,
        database_name: str,
        wise_data: WISEDataBase,
        status: str,
        service: str = "tap",
        load_from_bigdata_dir: bool = False
):
    """
    Plot all lightcurves of a given status.

    :param base_name: base name of the analysis
    :type base_name: str
    :param database_name: name of the database
    :type database_name: str
    :param wise_data: WISE data object
    :type wise_data: WISEDataBase
    :param status: status of the lightcurves to plot
    :type status: str
    :func:`timewise_sup.meta_analysis.ztf.get_ztf_lightcurve` or
    :func:`timewise_sup.meta_analysis.ztf.download_ztffp_per_status` before
    :param service: serice with which the lightcurves were downloaded by ``timewise`` (default: ``tap``)
    :type service: str, optional
    :param load_from_bigdata_dir: whether to load the lightcurves from the bigdata directory (default: ``False``)
    :type load_from_bigdata_dir: bool, optional
    """
    wise_lcs = get_ir_flux(
        base_name=base_name,
        database_name=database_name,
        wise_data=wise_data,
        status=status,
        service=service,
        load_from_bigdata_dir=load_from_bigdata_dir
    )

    for index, lc_dict in wise_lcs.items():
        logger.debug(f"index {index}")
        wise_lc = pd.DataFrame.from_dict(lc_dict)

        fig, ax = make_lightcurve_plot(wise_lc)

        d = plots_dir("baseline_subtracted_lightcurves", base_name)
        fn = os.path.join(d, status, f"{index}.pdf")

        d = os.path.dirname(fn)
        if not os.path.isdir(d):
            os.makedirs(d)

        logger.debug(f"saving under {fn}")
        fig.savefig(fn)
        plt.close()


def plot_single_lightcurve(
        base_name: str,
        database_name: str,
        wise_data: WISEDataBase,
        index: str,
        service: str = "tap",
        load_from_bigdata_dir: bool = False
) -> tuple[mpl.figure.Figure, mpl.axes.Axes]:
    """
    Plot a single lightcurve.

    :param base_name: base name of the analysis
    :type base_name: str
    :param database_name: name of the database
    :type database_name: str
    :param wise_data: WISE data object
    :type wise_data: WISEDataBase
    :param index: index of the lightcurve
    :type index: str
    :param service: service used to download the lightcurves, default "tap", passed to :func:`get_ir_flux_index`
    :type service: str, optional
    :param load_from_bigdata_dir: load from bigdata directory, defaults to False, passed to :func:`get_ir_flux_index`
    :type load_from_bigdata_dir: bool, optional
    :return: figure and axes
    :rtype: tuple[mpl.figure.Figure, mpl.axes.Axes]
    """

    wise_lc = pd.DataFrame.from_dict(get_ir_flux_index(
        base_name=base_name,
        database_name=database_name,
        wise_data=wise_data,
        index=index,
        service=service,
        load_from_bigdata_dir=load_from_bigdata_dir
    )[str(index)], orient="columns")

    return make_lightcurve_plot(wise_lc)


def make_lightcurve_plot(
        wise_lc: pd.DataFrame,
        ax: mpl.axes.Axes | None = None,
        x_key: str = "mean_mjd",
        y_key: str = flux_key,
        y_err_key: str = flux_err_key,
        ylabel: str = r"$\nu$ F$_{\nu}$ [erg s$^{-1}$ cm$^{-2}$]"
) -> tuple[mpl.figure.Figure, mpl.axes.Axes]:
    """
    Make a lightcurve plot.

    :param wise_lc: The WISE lightcurve
    :type wise_lc: pd.DataFrame
    :param ax: axes to plot on, defaults to creating a new figure
    :type ax: mpl.axes.Axes, optional
    :param x_key: key for the x axis, defaults to ``mean_mjd``
    :type x_key: str, optional
    :param y_key: key for the y axis, defaults to ``flux_key``
    :type y_key: str, optional
    :param y_err_key: key for the y error, defaults to ``flux_err_key``
    :type y_err_key: str, optional
    :param ylabel: label for the y axis, defaults to ``r"$\nu$ F$_{\nu}$ [erg s$^{-1}$ cm$^{-2}$]"``
    :type ylabel: str, optional
    :return: figure and axes
    :rtype: tuple[mpl.figure.Figure, mpl.axes.Axes]
    """

    fig, _ax = plt.subplots() if ax is None else (plt.gcf(), ax)  # type: mpl.figure.Figure, mpl.axes.Axes

    for b in ["W1", "W2"]:
        _ax.errorbar(
            wise_lc[x_key],
            wise_lc[f"{b}_{y_key}"],
            yerr=wise_lc[f"{b}_{y_err_key}"],
            marker='s',
            ms=3,
            color=bandcolors[b],
            label=f"WISE {b}",
            ls="",
            zorder=5,
            capsize=2
        )

    _ax.set_ylabel(ylabel)
    _ax.grid()
    _ax.legend()

    return fig, _ax



import logging
import pandas as pd
import matplotlib.pyplot as plt
from timewise import WISEDataDESYCluster
import numpy as np
import os

from timewise_sup.mongo import DatabaseConnector
from timewise_sup.plots import plots_dir


logger = logging.getLogger(__name__)


def get_red_chi2(
        wise_data: WISEDataDESYCluster,
) -> pd.DataFrame:
    logger.info(f"compiling red chi2 for {wise_data.base_name}")
    red_chi2_list = [
        wise_data.get_red_chi2(c, "_flux_density", use_bigdata_dir=True)
        for c in range(wise_data.n_chunks)
    ]
    logger.debug(f"loaded info for {len(red_chi2_list)} chunks")
    red_chi2_cit = {
        b: pd.concat([
            d[b][["chi2"]].rename(columns={"chi2": f"{b}_red_chi2"})
            for d in red_chi2_list
        ], axis=0)
        for b in wise_data.bands
    }
    red_chi2 = pd.concat(red_chi2_cit.values(), axis=1)
    red_chi2["pass_red_chi2_cut"] = (red_chi2["W1_red_chi2"] > 1) & (red_chi2["W2_red_chi2"] > 1)
    logger.info("done")
    return red_chi2


def get_res_info(
        base_name: str,
        database_name: str,
        wise_data: WISEDataDESYCluster
) -> pd.DataFrame:
    logger.info(f"getting result infos for {base_name}")
    database_connector = DatabaseConnector(base_name=base_name, database_name=database_name)
    status = database_connector.get_status(wise_data.parent_sample.df.index)
    desc = database_connector.get_t2_dust_echo_eval_descriptions(wise_data.parent_sample.df.index)
    red_chi2 = get_red_chi2(wise_data=wise_data)
    info = pd.concat([status, desc, red_chi2], axis=1)

    logger.info("done")
    return info


def make_autopct(values, relative=True):
    def my_autopct(pct):
        total = sum(values)
        val = int(round(pct*total/100.0))
        return f'{pct:.0f}' if relative else f"{val:.0f}"
    return my_autopct


def make_res_plot(
        base_name: str,
        database_name: str,
        wise_data: WISEDataDESYCluster,
        relative: bool = True
):
    logger.info(f"making result plot for {base_name}")
    info = get_res_info(base_name, database_name, wise_data)
    s = list(info.status[~info.status.isna()].unique())
    labels = ["Low $\chi^2$"] + s

    counts = [sum(~info.pass_red_chi2_cut)] + [
        np.sum(info.status[info.pass_red_chi2_cut & info.in_airflares] == si) for si in s]

    fig, ax = plt.subplots()

    ax.pie(
        counts, labels=labels,
        autopct=make_autopct(counts, relative=relative),
        pctdistance=0.8,
        labeldistance=1.08,
        colors=['grey', 'darkgrey', '#762A83', 'lightgrey', '#9970AB', '#C2A5CF'],
        textprops={'fontsize': 15},
        shadow=True,
        radius=2,
        startangle=40,
    )
    fig.tight_layout()

    this_plots_dir = plots_dir("test_data_res", base_name)

    fn = os.path.join(this_plots_dir, f"{base_name}_relative.pdf" if relative else f"{base_name}_total.pdf")
    logger.info(f"saving under {fn}")
    fig.savefig(fn)
    plt.close()



"""
Module to subtract the baseline from the lightcurves.

The baseline is determined by the T2BayesianBlocks algorithm and stored in the database. This module reads the
baseline values and subtracts them from the lightcurves. The resulting lightcurves are stored in a json file.

* :func:`apply_baseline_subtraction` reads the baseline values from the database and subtracts them from the lightcurves.
* :func:`get_baseline_subtracted_lightcurves` reads the lightcurves from the json file by status or creates it.
* :func:`get_lightcurves` reads the lightcurves from the json file by indices.
* :func:`get_single_lightcurve` reads the lightcurves from the json file by index.
"""

import json
import os.path
from functools import cache
import numpy as np
import logging
import tqdm

import pandas as pd

from timewise_sup.mongo import DatabaseConnector, Index, Status
from timewise_sup.environment import load_environment
from timewise import WiseDataByVisit, WISEDataDESYCluster


logger = logging.getLogger(__name__)


def apply_baseline_subtraction(
        wise_data: WiseDataByVisit,
        database_name: str,
        service: str = "tap",
        load_from_bigdata_dir: bool = True,
        status: Status | None = None,
        index: Index | None = None
) -> dict:
    """
    Apply baseline subtraction to the lightcurves. The baseline values are read from the database and subtracted from
    the lightcurves. The resulting lightcurves are returned as a dictionary. The dictionary keys are the IDs of the
    lightcurves. The values are the lightcurves themselves.

    :param wise_data: WISEDataByVisit object
    :type wise_data: WiseDataByVisit
    :param database_name: name of the database
    :type database_name: str
    :param service: :class:`timewise` service that was used to download the data, defaults to "tap"
    :type service: str, optional
    :param load_from_bigdata_dir: whether to load the data from the bigdata directory, defaults to True
    :type load_from_bigdata_dir: bool, optional
    :param status: status of the lightcurves to be processed, defaults to None
    :type status: Status, optional
    :param index: IDs of the lightcurves to be processed, defaults to None
    :type index: Index, optional
    :return: dictionary containing the baseline subtracted lightcurves
    :rtype: dict
    """

    logger.debug(f"status {status}")
    logger.debug(f"index {index}")

    database_connector = DatabaseConnector(base_name=wise_data.base_name, database_name=database_name)

    if status is not None:
        logger.debug(f"getting indices from status {status}, ignoring passed value {index}!")
        index = database_connector.get_ids(status=status)
    else:
        if index is None:
            raise ValueError("You must specify either of 'index' or 'status'!")

    logger.info("getting baseline values")

    # keys are stock IDs, values are baseline values
    baselines = database_connector.get_baselines(index=index)
    ids = np.array(list(baselines.keys()))
    logger.debug(f"got baselines for {len(ids)} objects")

    # find the chunk that holds the respective IDs
    chunk_numbers = np.array([wise_data._get_chunk_number(parent_sample_index=s) for s in ids])

    # set up empty directory for baseline corrected lightcurves
    diff_lcs = dict()

    # load the chunks one after the other
    logger.info(f"calculating difference flux for {len(ids)} objects in {len(np.unique(chunk_numbers))} chunks")
    for c in np.unique(chunk_numbers):
        logger.debug(f"chunk {c}")

        if load_from_bigdata_dir or isinstance(wise_data, WISEDataDESYCluster):
            timewise_data_product = wise_data.load_data_product(
                service=service,
                chunk_number=c,
                use_bigdata_dir=load_from_bigdata_dir
            )
        else:
            timewise_data_product = wise_data.load_data_product(
                service=service,
                chunk_number=c
            )

            # loop through IDs in this chunk
        for s in tqdm.tqdm(ids[chunk_numbers == c], desc=f"apply baseline subtraction to chunk {c}"):
            lc_in = pd.DataFrame.from_dict(timewise_data_product[str(s)]["timewise_lightcurve"])
            lc_out = lc_in[["mean_mjd"]].copy()

            for b in ["W1", "W2"]:

                # loop over flux and flux errors
                f_key = WiseDataByVisit.mean_key + WiseDataByVisit.flux_density_key_ext
                ferr_key = WiseDataByVisit.flux_density_key_ext + WiseDataByVisit.rms_key

                try:
                    # f is a dict with keys: index and values: flux_densities
                    f = lc_in[b + f_key]
                    baseline = baselines[s][f"{b}_baseline"]
                    baseline_err = baselines[s][f"{b}_baseline_sigma"]

                    if baseline is None:
                        logger.warning(f"baseline for {s} {b} is None, skipping!")
                        continue

                    f_diff = {k: v - baseline for k, v in f.items()}
                    f_diff_key = "_diff" + f_key
                    lc_out.loc[list(f_diff.keys()), b + f_diff_key] = list(f_diff.values())

                    ferr_diff_key = "_diff" + ferr_key
                    lc_out[b + ferr_diff_key] = np.sqrt(lc_in[b + ferr_key].copy() ** 2 + baseline_err ** 2)

                except KeyError as e:
                    logger.error(
                        f"parent sample index {s}: {e}. "
                        f"LC is \n{json.dumps(lc_out, indent=4)}"
                    )
                    raise KeyError

            # save lightcurve containing baseline corrected flux
            diff_lcs[str(s)] = lc_out.to_dict()

    logger.info("done")

    return diff_lcs


def get_baseline_subtracted_lightcurves(
        base_name: str,
        database_name: str,
        wise_data: WiseDataByVisit,
        status: Status,
        force_new: bool = False,
        service: str = "tap",
        load_from_bigdata_dir: bool = True,
) -> dict:
    """
    Get the baseline subtracted lightcurves from the json file buy status. If the file does not exist, call the function
    :func:`apply_baseline_subtraction` to create it. The resulting lightcurves are returned as a dictionary. The
    dictionary keys are the IDs of the lightcurves. The values are the lightcurves themselves.

    :param base_name: name of the base directory
    :type base_name: str
    :param database_name: name of the database
    :type database_name: str
    :param wise_data: WISEDataByVisit object
    :type wise_data: WiseDataByVisit
    :param status: status of the lightcurves to be processed
    :type status: Status
    :param force_new: whether to force the creation of a new file, defaults to False
    :type force_new: bool, optional
    :param service: :class:`timewise` service that was used to download the data, defaults to "tap"
    :type service: str, optional
    :param load_from_bigdata_dir: whether to load the data from the bigdata directory, defaults to True
    :type load_from_bigdata_dir: bool, optional
    :return: dictionary containing the baseline subtracted lightcurves
    :rtype: dict
    """
    logger.info(f"getting baseline subtracted lightcurves for status {status}")
    tsup_data_dir = load_environment("TIMEWISE_SUP_DATA")
    fn = os.path.join(tsup_data_dir, base_name, f"diff_lcs_status{status}.json")

    if (not os.path.isfile(fn)) or force_new:
        logger.info(f"No file {fn}.")
        logger.info("Making baseline subtracted lightcurves")
        diff_lcs = apply_baseline_subtraction(
            wise_data=wise_data,
            database_name=database_name,
            service=service,
            load_from_bigdata_dir=load_from_bigdata_dir,
            status=status
        )

        d = os.path.dirname(fn)
        if not os.path.isdir(d):
            os.makedirs(d)

        logger.debug(f"saving under {fn}")
        with open(fn, "w") as f:
            json.dump(diff_lcs, f)

    else:
        logger.debug(f"loading {fn}")
        with open(fn, "r") as f:
            diff_lcs = json.load(f)

    return diff_lcs


def get_lightcurves(
        base_name: str,
        database_name: str,
        wise_data: WiseDataByVisit,
        index: Index,
        service: str = "tap",
        load_from_bigdata_dir: bool = True,
) -> dict:
    """
    Get the baseline subtracted lightcurves from the json file by index. If the file does not exist, call the function
    :func:`get_baseline_subtracted_lightcurves` to create it. The resulting lightcurves are returned as a dictionary.
    The dictionary keys are the IDs of the lightcurves. The values are the lightcurves themselves.

    :param base_name: name of the base directory
    :type base_name: str
    :param database_name: name of the database
    :type database_name: str
    :param wise_data: WISEDataByVisit object
    :type wise_data: WiseDataByVisit
    :param index: IDs of the lightcurves to be processed
    :type index: Index
    :param service: :class:`timewise` service that was used to download the data, defaults to "tap"
    :type service: str, optional
    :param load_from_bigdata_dir: whether to load the data from the bigdata directory, defaults to True
    :type load_from_bigdata_dir: bool, optional
    :return: dictionary containing the baseline subtracted lightcurves
    :rtype: dict
    """
    indices = list(np.atleast_1d(index))
    logger.info(f"getting lightcurves {len(indices)} objects")
    status = DatabaseConnector(base_name=base_name, database_name=database_name).get_status(index=tuple(indices))
    logger.debug(f"found {len(status['status'].unique())} statuses")
    lcs = dict()
    for s in status["status"].unique():
        iids = status.index[status["status"] == s]
        slcs = get_baseline_subtracted_lightcurves(
            base_name=base_name,
            database_name=database_name,
            wise_data=wise_data,
            status=s,
            service=service,
            load_from_bigdata_dir=load_from_bigdata_dir
        )

        for i in iids:
            lcs[str(i)] = slcs[i]

    logger.debug(f"returning {len(lcs)} lightcurves")
    return lcs


@cache
def get_single_lightcurve(
        base_name: str,
        database_name: str,
        wise_data: WiseDataByVisit,
        index: str,
        service: str = "tap",
        load_from_bigdata_dir: bool = True,
) -> pd.DataFrame:
    """
    Get the baseline subtracted lightcurves from the json file by index. If the file does not exist, call the function
    :func:`get_lightcurves` to create it.

    :param base_name: name of the base directory
    :type base_name: str
    :param database_name: name of the database
    :type database_name: str
    :param wise_data: WISEDataByVisit object
    :type wise_data: WiseDataByVisit
    :param index: ID of the lightcurve to be processed
    :type index: str
    :param service: :class:`timewise` service that was used to download the data, defaults to "tap"
    :type service: str, optional
    :param load_from_bigdata_dir: whether to load the data from the bigdata directory, defaults to True
    :type load_from_bigdata_dir: bool, optional
    :return: baseline subtracted lightcurve
    :rtype: pandas.DataFrame
    """
    lcs = get_lightcurves(
        base_name=base_name,
        database_name=database_name,
        wise_data=wise_data,
        index=index,
        service=service,
        load_from_bigdata_dir=load_from_bigdata_dir
    )
    return pd.DataFrame.from_dict(lcs[str(index)])



"""
Module to interface with the :class:`MongoDB` database.

* :class:`DatabaseConnector`:
    The main class to interface with the database. It is initialized with the :attr:`base_name` and
    :attr:`database_name` and provides methods to get data from the database that was produced by :code:`AMPEL`.
"""

import logging
import os
import json
import pandas as pd
import numpy as np
from tqdm import tqdm
from pymongo import MongoClient, collection, database
from collections.abc import Sequence
from pydantic import BaseModel
from pathlib import Path

from timewise_sup.environment import load_environment


logger = logging.getLogger(__name__)


jd_offset = -2400000.5


################################################################################
# -----------------------         utilities            ----------------------- #
# --------------------------------  START  ----------------------------------- #
#


Index = Sequence[str | int] | str | int
Status = str | Sequence[str]


def as_list(
        index: Index
) -> list:
    """
    Creates a list from input and transforms np.int64 to int

    :param index: objects to transform to list
    :type index: Index
    :return: the list
    :rtype: list
    """
    indices = list(np.atleast_1d(index))

    for i, ii in enumerate(indices):
        if isinstance(ii, np.int64):
            indices[i] = int(ii)

    return indices


def chunks(
        the_list: Sequence,
        length: int
):
    """
    Yield sequential chunks from l with fixed length

    :param the_list: sequence to be split into chunks
    :type the_list: Sequence
    :param length: length of isngle chunk
    :type length: int
    :return: generator creating the chunks
    :rtype: generator
    """
    number_of_chunks = int(np.ceil(len(the_list) / length))
    logger.debug(f"splitting {len(the_list)} into {number_of_chunks} chunks of length {length}")
    d, r = divmod(len(the_list), number_of_chunks)
    for i in range(number_of_chunks):
        logger.debug(f"chunk {i}")
        si = (d+1)*(i if i < r else r) + d*(0 if i < r else i - r)
        yield the_list[si:si+(d+1 if i < r else d)]


#
# -----------------------         utilities            ----------------------- #
# --------------------------------   END   ----------------------------------- #
################################################################################


################################################################################
# -----------------------      DatabaseConnector       ----------------------- #
# --------------------------------  START  ----------------------------------- #
#


class DatabaseConnector(BaseModel):
    """
    Class to connect to `MongoDB`

    :param base_name: The name given to the :class:`WISEData`
    :type base_name: str
    :param database_name: The name of the database in the :class:`MongoDB`
    :type database_name: str
    :param database: The database object from :class:`pymongo`
    :type database: :class:`pymongo.database.Database`
    :param t2collection: The collection object representing the tier 2 collection as used by :class:`AMPEL`
    :type t2collection: :class:`pymongo.collection.Collection`
    """

    base_name: str
    database_name: str
    database: database.Database
    t2collection: collection.Collection

    class Config:
        arbitrary_types_allowed = True

    def __init__(self, **kwargs):
        # get the client, database and t2 collections
        client: MongoClient = DatabaseConnector.connect_to_db()
        _database_name = kwargs["database_name"]

        if _database_name not in client.list_database_names():
            logger.warning(f"No database with name {_database_name} registered in client!")

        _database = client[_database_name]
        kwargs["database"] = _database
        kwargs["t2collection"] = _database["t2"]

        super().__init__(**kwargs)

    @property
    def cache_dir(self) -> Path:
        """Directory used to store cache"""
        tsup_data = Path(load_environment("TIMEWISE_SUP_DATA"))
        return tsup_data / self.base_name / self.database_name

    ################################################################################
    # -----------------------  interface with MongoDB      ----------------------- #
    # --------------------------------  START  ----------------------------------- #
    #

    @staticmethod
    def connect_to_db() -> MongoClient:
        """Return the :class:`MongoClient`"""
        mongodb_port = load_environment("TIMEWISE_SUP_MONGODB_PORT")
        logger.debug(f"connecting to MongoDB at {mongodb_port}")
        client: MongoClient = MongoClient(f"mongodb://localhost:{mongodb_port}/")
        logger.debug("connected")
        return client

    def get_dataframe(
            self,
            collection_name: str,
            field_name_lists: list[list[str | int]],
            filter: dict | None = None
    ) -> pd.DataFrame:
        """
        Get fields from a collection in the database and return as :class:`pandas.DataFrame`.

        :param collection_name: Name of the collection
        :type collection_name: str
        :param field_name_lists: The "path" to the value in the collection
        :type field_name_lists: list of lists
        :param filter: filter to apply to the entries
        :type filter: dict
        :return: the selected data
        :rtype: pandas.DataFrame
        """
        logger.info(f"making dataframe from {field_name_lists} of {self.database_name}.{collection_name}")
        col = self.database[collection_name]  # type: collection.Collection
        filter = dict() if filter is None else filter

        res: dict[str, list] = {f"field{i}": list() for i in range(len(field_name_lists))}
        for entry in col.find(filter):
            for i_field_names, field_names in enumerate(field_name_lists):

                # loop through the list of field names / indices to dig down to the value
                val = entry
                for field_name in field_names:
                    if field_name not in val:
                        val = np.nan
                        break
                    val = val[field_name]

                res[f"field{i_field_names}"].append(val)

        return pd.DataFrame.from_dict(res)

    def drop_stocks(self, stocks: Index, tiers: list[int] | None = None):
        """Drop lightcurves and related results from the database"""
        indices = as_list(stocks)
        _tiers = [0, 1, 2] if tiers is None else tiers
        logger.debug(f"dropping {len(indices)} indices of {self.base_name} from {self.database_name} (tiers {_tiers})")

        f = {"stock": {"$in": indices}}

        if 0 in _tiers:
            t0 = self.database["t0"]
            logger.info("dropping datapoints")
            for index in tqdm(indices, desc="dropping datapoints"):
                t0.delete_many({"stock": int(index)})

        if 1 in _tiers:
            t1 = self.database["t1"]
            logger.info("dropping lightcurves")
            t1.delete_many(f)

        if 2 in _tiers:
            logger.info("dropping T2 results")
            self.t2collection.delete_many(f)

    #
    # -----------------------  interface with MongoDB      ----------------------- #
    # --------------------------------   END   ----------------------------------- #
    ################################################################################

    ################################################################################
    # ------------------     get T2DustEchoEval results       -------------------- #
    # --------------------------------  START  ----------------------------------- #
    #

    def get_status(
            self,
            index: Index
    ) -> pd.DataFrame:
        """
        Get the status of objects as determined by the pipline.

        :param index: The index of the objects in the :class:`timewise.ParentSample`
        :type index: Index
        :return: Status of the objects
        :type: pandas.DataFrame
        """
        indices = as_list(index)
        logger.debug(f"getting status for {len(indices)} IDs ({self.base_name} in {self.database_name})")

        columns = {
            "stock": 1,
            "body.status": 1
        }

        i = None
        status = dict()
        for chunked_indices in chunks(indices, int(1e6)):
            dust_echo_filter = {
                "unit": "T2DustEchoEval",
                "code": 0,
                "stock": {"$in": chunked_indices}
            }
            for i, x in enumerate(self.t2collection.find(dust_echo_filter, columns)):
                status[str(x["stock"])] = {"status": x['body'][-1]['status']}

        if len(status) == 0:
            return pd.DataFrame({"status": []})

        return pd.DataFrame.from_dict(status, orient="index")

    def get_unique_statuses(self) -> list:
        """Returns a :class:`list` of all unique statuses"""
        filter = {
            "unit": "T2DustEchoEval",
            "code": 0
        }
        statuses = self.t2collection.distinct("body.status", filter)
        return statuses

    def _calculate_ids(self, status: str) -> list:
        """
        Finds the IDs for a status

        :param status: The status for which the IDs should be calculated for
        :type status: str
        :return: The IDs
        :rtype: list
        """
        logger.debug(f"calculating all IDs for status {status} ({self.base_name} in {self.database_name})")

        filter = {
            "unit": "T2DustEchoEval",
            "body.status": status,
            "code": 0
        }

        columns = {"stock": 1, "body.status": 1}
        stocks = list()
        n_wrong = 0
        for i in tqdm(self.t2collection.find(filter, columns), desc=f"getting IDs for status {status}"):
            stock = i["stock"]
            if i["body"][-1]["status"] != status:
                n_wrong += 1
            else:
                stocks.append(stock)

        stocks = as_list(np.unique(stocks))
        logger.debug(f"found {len(stocks)} stocks, {n_wrong} had matching status from previous run.")
        return stocks

    def get_ids(self, status: Status) -> list:
        """
        Get the IDs for a status. When called for the first time, caches the results to a file
        that will be loaded subsequently.

        :param status: The status
        :type status: Status
        :return: The IDs
        :rtype: list
        """
        logger.debug(f"getting all IDs for status {status} ({self.base_name} in {self.database_name})")

        ids_cache_dir = self.cache_dir / "ids"
        ids_cache_dir.mkdir(parents=True, exist_ok=True)

        stocks = list()

        for istatus in np.unique(status).astype(str):
            fn = ids_cache_dir / f"status{istatus.replace(' ', '_')}.json"

            if not os.path.isfile(fn):
                istocks = self._calculate_ids(istatus)
                logger.debug(f"caching to {fn}")
                with open(fn, "w") as f:
                    json.dump(istocks, f)

            else:
                logger.debug(f"loading cache from {fn}")
                with open(fn, "r") as f:
                    istocks = json.load(f)

                logger.debug(f"got {len(istocks)} ids")

            stocks.extend(istocks)

        return stocks

    def get_excess_mjd(self, index: Index) -> pd.DataFrame:
        """
        Get the excess time for the given index

        :param index: The wanted index
        :type index: Index
        :return: Modified Julian Days of the excess, per filter, start, end and if flare ended
        :rtype: dict
        """
        indices = as_list(index)
        logger.debug(f"getting excess time for {len(indices)} IDs ({self.base_name} in {self.database_name})")

        excess_mjds = dict()

        for chunked_indices in chunks(indices, int(1e6)):
            bayesian_blocks_filter = {
                "unit": "T2BayesianBlocks",
                "stock": {"$in": chunked_indices},
                "code": 0
            }

            for i, x in enumerate(self.t2collection.find(bayesian_blocks_filter)):
                i_excess_mjds = dict()
                for f in ["W1", "W2"]:
                    excess_mags = np.array(x["body"][-1][f"Wise_{f}"]["max_mag_excess_region"])
                    excess_jds = x["body"][-1][f"Wise_{f}"]["jd_excess_regions"]
                    max_excess_mjds = np.array(excess_jds)[np.argmax(excess_mags)] + jd_offset

                    baseline_mjds = np.array(x["body"][-1][f"Wise_{f}"]["jd_baseline_regions"]).flatten() + jd_offset
                    excess_ended = max(baseline_mjds) > max(max_excess_mjds)

                    i_excess_mjds[f"{f}_excess_start_mjd"] = min(max_excess_mjds)
                    i_excess_mjds[f"{f}_excess_end_mjd"] = max(max_excess_mjds)
                    i_excess_mjds[f"{f}_flare_ended"] = excess_ended

                excess_mjds[str(x["stock"])] = i_excess_mjds

        return pd.DataFrame.from_dict(excess_mjds, orient="index")

    def get_baselines(self, index: Index) -> dict:
        """
        Returns the baseline for the lightcurves.

        :param index: The index of the lightcurves
        :type index: Index
        :returns: The value of the baseline flux and the :math:`1\sigma` uncertainty per filter
        :rtype: dict
        """
        baseline = dict()

        indices = as_list(index)
        logger.debug(f"reading baseline values for {len(indices)} objects")

        for chunked_indices in chunks(indices, int(5e5)):

            dust_echo_filter = {
                "unit": "T2DustEchoEval",
                "stock": {"$in": chunked_indices},
                "code": 0
            }

            for x in self.t2collection.find(
                    dust_echo_filter,
                    {"stock": 1}
            ):
                i_baseline = dict()
                for y in self.t2collection.find(
                        {"unit": "T2BayesianBlocks", "stock": x["stock"], "code": 0},
                        {f"body.Wise_{b}.baseline{s}": 1 for b in ["W1", "W2"] for s in ["", "_sigma"]},
                        limit=1
                ):
                    for b in ["W1", "W2"]:
                        i_baseline[f"{b}_baseline"] = y["body"][-1][f"Wise_{b}"]["baseline"]
                        i_baseline[f"{b}_baseline_sigma"] = y["body"][-1][f"Wise_{b}"]["baseline_sigma"]

                baseline[str(x["stock"])] = i_baseline

        return baseline

    def _calculate_t2_dust_echo_eval_descriptions(self, index: Index) -> pd.DataFrame:
        """
        Get the description for the lightcurves

        :param index: The index of the lightcurves
        :type index: Index
        :return: Description per lightcurves, separated by ", "
        :rtype: dict
        """
        indices = as_list(index)
        logger.debug(f"getting description for {len(indices)} objects of {self.base_name} in {self.database_name}")

        values = dict()

        for chunked_indices in chunks(indices, int(5e5)):
            filter = {
                "code": 0,
                "unit": "T2DustEchoEval",
                "stock": {"$in": chunked_indices}
            }

            for lc in self.t2collection.find(filter):
                value = dict()

                try:
                    value[f"description"] = ", ".join(lc["body"][-1]["description"])
                except KeyError as e:
                    raise KeyError(f"{lc}: {e}")

                values[str(lc["stock"])] = value

        logger.debug(f"returning {len(values)} results")

        return pd.DataFrame.from_dict(values, orient="index")

    def get_t2_dust_echo_eval_descriptions(self, indices: Index) -> pd.DataFrame:
        """
        Get the description of the lightcurves. Caches result to file when first called.

        :param indices: The index of the lightcurves
        :type indices: Index
        :return: Description per lightcurves, separated by ", "
        :rtype: dict
        """
        fn = self.cache_dir / "T2DustEchoEvalDescriptions.csv"
        fn.parents[0].mkdir(parents=True, exist_ok=True)

        if not os.path.isfile(fn):
            logger.debug(f"no file {fn}")
            desc = self._calculate_t2_dust_echo_eval_descriptions(indices)
            logger.debug(f"saving to {fn}")
            desc.to_csv(fn)

        else:
            logger.debug(f"loading {fn}")
            desc = pd.read_csv(fn, index_col=0)

        indices_series = pd.Series(indices)
        indices_present = indices_series.astype(str).isin(desc.index.astype(str))
        if np.any(~indices_present):
            logger.debug(f"{np.sum(~indices_present)} indices not found. calculating")
            desc_suplement = self._calculate_t2_dust_echo_eval_descriptions(indices_series[~indices_present])
            desc = pd.concat([desc, desc_suplement])
            logger.debug(f"saving {len(desc)} descriptions to {fn}")
            desc.to_csv(fn)

        return desc.loc[indices_series.astype(type(desc.index[0]))]

    def get_agn_variability_stocks(self) -> list:
        logger.debug("getting AGN stock IDs based on variability")

        filter = {
            '$and': [
                {"code": 0},
                {"unit": "T2DustEchoEval"},
                # {"body": {"$elemMatch": {"description": "Excess region exists"}}},
                {"body.description": {"$elemMatch": {"$ne": "Baseline only"}}},
                {"body.description": {"$elemMatch": {"$ne": "Stage transition"}}},
                {"body.status": {"$ne": "1"}}
                ]
        }

        cols = {"stock": 1}

        stocks = list()
        for lc in self.t2collection.find(filter, cols):
            stocks.append(lc["stock"])

        logger.debug(f"returning {len(stocks)} stock IDs")
        return stocks

    def get_t2_dust_echo_eval_values(self, index: Index, value: str) -> pd.DataFrame:
        """
        Get a value per filter and lightcurve as produced by the :class:`AMPEL` :class:`T2DustEchoEval`

        :param index: The index of the lightcurves
        :type index: Index
        :param value: The key of the value
        :type value: str
        :return: The value per lightcurves and filter
        :rtype: pandas.DataFrame
        """

        indices = as_list(index)
        logger.debug(f"getting {value} for {len(indices)} objects of {self.base_name} in {self.database_name}")

        filter = {
            "unit": "T2DustEchoEval",
            "stock": {"$in": as_list(index)},
            "code": 0
        }

        n = self.t2collection.count_documents(filter)
        logger.debug(f"found {n} documents")

        values = dict()

        for lc in self.t2collection.find(filter):
            ivalue = dict()

            for i, b in enumerate(["W1", "W2"]):
                try:
                    ivalue[f"{value}_{b}"] = lc["body"][-1]["values"][value][i]
                except KeyError as e:
                    raise KeyError(f"{lc}: {e}")
                except IndexError as e:
                    raise IndexError(f"{lc}: {e}")

            values[str(lc["stock"])] = ivalue

        logger.debug(f"returning {len(values)} results")

        return pd.DataFrame.from_dict(values, orient="index")

    def get_peak_time_jd(self, index: Index) -> pd.DataFrame:
        """Get the peak time in Julian Days"""
        return self.get_t2_dust_echo_eval_values(index, "max_mag_jd")

    def get_excess_end_jd(self, index: Index) -> pd.DataFrame:
        """Get the start of the excess in Julian Days"""
        return self.get_t2_dust_echo_eval_values(index, "excess_jd")

    def get_dust_echo_strength(self, index: Index) -> pd.DataFrame:
        """Get the strength of the dust echo"""
        return self.get_t2_dust_echo_eval_values(index, "strength_sjoert")

    def get_fade_time(self, index: Index) -> pd.DataFrame:
        """Get the fade time"""
        return self.get_t2_dust_echo_eval_values(index, "e_fade")

    def get_rise_time(self, index: Index) -> pd.DataFrame:
        """Get the rise time"""
        return self.get_t2_dust_echo_eval_values(index, "e_rise")

    #
    # ------------------     get T2DustEchoEval results       -------------------- #
    # --------------------------------   END   ----------------------------------- #
    ################################################################################

    ############################################################################################
    # -------------     get T2CatalogMatch and T2DigestRedshifts results       --------------- #
    # -------------------------------------  START  ------------------------------------------ #
    #

    def get_catalog_matches(self, index: Index, code: int = 0) -> dict:
        """
        Get the matches as determined by the :class:`AMPEL` :class:`T2CatalogMatch`

        :param index: Index of the Lightcurves
        :type index: Index
        :param code: The document code to query for
        :type code: int
        :return: All matches in the catalogues
        :rtype: dict
        """
        logger.debug(f"getting catalog match info for {self. base_name} in {self.database_name}")
        indices = as_list(index)
        logger.debug(f"getting catalog match info for {len(indices)} objects")

        matches = dict()

        for chunked_indices in chunks(indices, 1000000):

            filter = {
                "code": code,
                "stock": {"$in": chunked_indices},
                "unit": "T2CatalogMatch"
            }

            for i in self.t2collection.find(filter):
                matches[i["stock"]] = i["body"][-1]

        return matches

    def get_redshift(self, index: Index) -> pd.DataFrame:
        """
        Get the redshifts as determined by the :class:`AMPEL` :class:`T2DigestRedshifts`

        :param index: Index of the lightcurves
        :type index: Index
        :return: The redshift, distance to the match and the uncertainty associated to the redshift measurement method
        :rtype: pandas.DataFrame
        """
        filter = {
            "unit": "T2DigestRedshifts",
            "stock": {"$in": as_list(index)},
        }

        n = self.t2collection.count_documents(filter)
        logger.debug(f"found {n} documents")

        redshift = dict()

        for lc in self.t2collection.find(filter):
            try:
                if "ampel_z" in lc["body"][-1]:
                    d = {k: lc["body"][-1][k] for k in ["ampel_z", "ampel_dist", "group_z_precision"]}
                    redshift[str(lc["stock"])] = d
                else:
                    redshift[str(lc["stock"])] = {"ampel_z": np.nan, "ampel_dist": np.nan, "group_z_precision": np.nan}
            except KeyError as e:
                raise KeyError(f"{lc}: {e}")

        logger.debug(f"returning {len(redshift)} results")

        return pd.DataFrame.from_dict(redshift, orient="index")

    #
    # -------------     get T2CatalgueMatch and T2DigestRedshifts results       --------------- #
    # --------------------------------------   END   ------------------------------------------ #
    #############################################################################################

#
# -----------------------      DatabaseConnector       ----------------------- #
# --------------------------------   END   ----------------------------------- #
################################################################################



"""
This module contains the function to get the number of objects per rejection reason. The rejection reason is given in
the description of the ``AMPEL`` ``T2DustEchoEval`` unit output.

* :func:`get_rejection_reason` gets the number of objects per rejection reason for the whole sample
* :func:`calculate_rejection_reasons` calculates the rejection reasons for the given index
* :func:`get_individual_rejection_reasons` gets the individual rejection reasons for the given index and caches to disk
"""

import logging
import json
import numpy as np
import pandas as pd

from timewise_sup.mongo import DatabaseConnector, Index


logger = logging.getLogger(__name__)


def get_rejected_ids(
        base_name: str,
        database_name: str,
):
    """
    Get the ids of the rejected objects. The rejection reason is given in the description of the
    ``AMPEL`` ``T2DustEchoEval`` unit output.

    :param base_name: base name of the wise data
    :type base_name: str
    :param database_name: name of the database
    :type database_name: str
    :return: ids of the rejected objects
    :rtype: list
    """
    logger.info(f"getting rejected ids for {base_name}")
    rejected = (
        "No further investigation",
        "1_maybe_interesting",
        "2_maybe_interesting",
        "3", "3_maybe_interesting",
        "4", "4_maybe_interesting"
    )
    database_connector = DatabaseConnector(base_name=base_name, database_name=database_name)
    ids = database_connector.get_ids(rejected)
    return ids


def get_rejection_reason(
        base_name: str,
        database_name: str
) -> dict:
    """
    Get the number of objects per rejection reason. The rejection reason is given in the description of the
    ``AMPEL`` ``T2DustEchoEval`` unit output.

    :param base_name: base name of the wise data
    :type base_name: str
    :param database_name: name of the database
    :type database_name: str
    :return: dictionary with the rejection reason as key and the number of objects as value
    :rtype: dict
    """
    logger.info(f"getting rejection reason for {base_name}")
    ids = get_rejected_ids(base_name, database_name)
    database_connector = DatabaseConnector(base_name=base_name, database_name=database_name)
    desc = database_connector.get_t2_dust_echo_eval_descriptions(ids)
    desc_counts = desc.description.value_counts()  # type: pd.Series

    # when there is a gap in the lightcurve before a potential excess, the status will be "_maybe_interesting"
    # but can have the accepted description. We can replace it with the more meaningful "gap"
    ac_str = 'Baseline before excess region, Excess region exists, Baseline before excess region, Excess region exists'
    desc_counts.rename(index={ac_str: "gap"}, inplace=True)

    individual_reasons = dict()
    for ir, n in desc_counts.items():
        # descriptions are given per band
        # here we make sure to only count them once
        already_counted = []
        for irsi in ir.split(", "):
            if irsi in already_counted:
                continue
            if irsi not in individual_reasons:
                individual_reasons[irsi] = n
            else:
                individual_reasons[irsi] += n
            already_counted.append(irsi)

    # sort by number
    individual_reasons_df = pd.DataFrame.from_dict(individual_reasons, orient="index").sort_values(0, ascending=False)

    # find individual reasons in descriptions
    accept = ['Baseline before excess region', 'Excess region exists']
    masks: dict[str, bool] = {
        rt: desc_counts.index.str.contains(rt)
        for rt in individual_reasons_df.index if rt not in accept
    }

    # find individual reasons that exclude the object with descending order of appearance
    masks_exclusive = {rt: im & ~np.logical_or.reduce(list(masks.values())[:i])
                       for i, (rt, im) in enumerate(masks.items())}

    # find the corresponding numbers of objects
    numbers_exclusive = {rt: int(desc_counts[m].sum()) for rt, m in masks_exclusive.items()}
    logger.debug(json.dumps(numbers_exclusive, indent=4))

    return numbers_exclusive


def calculate_rejection_reasons(
        base_name: str,
        database_name: str,
        index: Index
) -> pd.DataFrame:
    """
    Calculate the rejection reasons for the given index.

    :param base_name: base name of the wise data
    :type base_name: str
    :param database_name: name of the database
    :type database_name: str
    :param index: index of the objects
    :type index: Index
    :return: dataframe with the rejection reasons
    :rtype: pd.DataFrame
    """
    connector = DatabaseConnector(base_name=base_name, database_name=database_name)
    _index = pd.Index(index)
    info = pd.DataFrame(index=_index, columns=["rejection_reason"], dtype=str)

    # find rejected ids
    rejected_ids = get_rejected_ids(base_name, database_name)

    # find indices that have not been run, probably because of chi2 cut
    accepted_index = pd.Index(connector.get_ids(("1", "2")))
    info.loc[_index[_index.isin(index)], "rejection_reason"] = "accepted"

    # find indices that have not been run, probably because of chi2 cut
    not_run_index = _index[~_index.isin(accepted_index) & ~_index.isin(rejected_ids)]
    info.loc[not_run_index, "rejection_reason"] = "not run"

    # get the descriptions for the rejected indices
    desc = connector.get_t2_dust_echo_eval_descriptions(_index[_index.isin(rejected_ids)])

    # where the description suggests acception but the objects was rejected, we have a gap in the lightcurve
    ac_str = 'Baseline before excess region, Excess region exists, Baseline before excess region, Excess region exists'
    gap_mask = desc.description.str.contains(ac_str) & desc.index.isin(rejected_ids)
    desc.loc[gap_mask, "description"] = "gap"

    # set the rejection reason for the rejected indices
    global_reasons_dict = get_rejection_reason(base_name, database_name)
    global_reasons = pd.DataFrame.from_dict(global_reasons_dict, orient="index").sort_values(0, ascending=True)
    for r in global_reasons.index:
        i = desc.index[desc.description.str.contains(r)]
        info.loc[_index.astype(str).isin(i.astype(str)), "rejection_reason"] = r

    return info


def get_individual_rejection_reasons(
        base_name: str,
        database_name: str,
        index: Index
) -> pd.DataFrame:
    """
    Get the individual rejection reasons for the given index. If the individual rejection reasons are not yet
    calculated, they will be calculated and saved.

    :param base_name: base name of the wise data
    :type base_name: str
    :param database_name: name of the database
    :type database_name: str
    :param index: index of the objects
    :type index: Index
    :return: dataframe with the individual rejection reasons
    :rtype: pd.DataFrame
    """

    connector = DatabaseConnector(base_name=base_name, database_name=database_name)
    filename = connector.cache_dir / f"individual_rejection_reasons.csv"

    if filename.is_file():
        logger.debug(f"loading individual rejection reasons from {filename}")
        individual_reasons = pd.read_csv(filename, index_col=0)
    else:
        individual_reasons = pd.DataFrame()

    _index = pd.Index(index)
    is_calculated_mask = _index.isin(individual_reasons.index)

    if any(~is_calculated_mask):
        logger.debug(f"calculating individual rejection reasons for {sum(~is_calculated_mask)} objects of {base_name} "
                     f"in {database_name}")
        new_reasons = calculate_rejection_reasons(base_name, database_name, _index[~is_calculated_mask])
        individual_reasons = pd.concat([individual_reasons, new_reasons], axis=0)
        logger.debug(f"saving individual rejection reasons to {filename}")
        individual_reasons.to_csv(filename)

    return individual_reasons.loc[_index]




"""
Makes color-color plots for the WISE data.

* :func:`color_plot` makes a color-color plot given the wise data and a list of color keys in the parent sample.
* :func:`agn_box_plot` makes a color-color plot in the W1-W2 - W2-W3 plane and indicates the `AGN-Wedge`
"""

import logging
import os.path

import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np
import corner
import pandas as pd
from timewise import WISEDataDESYCluster
import numpy.typing as npt

from timewise_sup.plots import plots_dir
from timewise_sup.meta_analysis.agn import add_agn_box, get_agn_mask_from_color
from timewise_sup.mongo import DatabaseConnector, Status


logger = logging.getLogger(__name__)


def color_plot(
        base_name: str,
        database_name: str,
        wise_data: WISEDataDESYCluster,
        color_keys: list[str],
        status: Status | None = None,
        colors: list[str] | None = None,
        level_sigma: list[float] | None = None,
        data: pd.DataFrame | None = None,
) -> tuple[mpl.figure.Figure, npt.NDArray]:
    """
    Makes a color-color plot given the wise data and a list of color keys in the parent sample.

    :param base_name: base name for storage directories
    :type base_name: str
    :param database_name: name of the database
    :type database_name: str
    :param wise_data: instance of WISEDataDESYCluster
    :type wise_data: WISEDataDESYCluster
    :param color_keys: list of color keys in the parent sample (or ``data``)
    :type color_keys: list
    :param status: statuses to include in the plot, defaults to ``["1", "2"]``
    :type status: list, optional
    :param colors: colors to use for the different statuses, defaults to ``["k", "grey"]``
    :type colors: list, optional
    :param level_sigma: sigma levels for the contours, defaults to ``[1, 2, 5]``
    :type level_sigma: list, optional
    :param data: data to use for the plot, defaults to ``wise_data.parent_sample.df``
    :type data: pd.DataFrame, optional
    :return: figure and axes
    :rtype: tuple
    """

    if status is None:
        status = ["1", "2"]

    if colors is None:
        colors = ["k", "grey"]

    if level_sigma is None:
        level_sigma = [1, 2, 5]

    if data is None:
        data = wise_data.parent_sample.df
        data.set_index(data.index.astype(str), inplace=True)

    fig, axs = plt.subplots(
        ncols=len(color_keys),
        nrows=len(color_keys),
        figsize=[6, 6],
        sharex='all',
        sharey='all',
        gridspec_kw={"wspace": 0.1, "hspace": 0.1}
    )

    twinx = [None] * len(color_keys)

    for i_status, c in zip(status, colors):

        ids = DatabaseConnector(base_name=base_name, database_name=database_name).get_ids(i_status)

        for i in range(len(color_keys)):
            for j in range(i + 1, len(color_keys)):
                x = np.array(data.loc[list(ids), color_keys[i]])
                y = np.array(data.loc[list(ids), color_keys[j]])

                logger.debug(f"making corner plot: {color_keys[i]} ({len(x)}) - {color_keys[j]} ({len(y)})")
                nan_m = np.isnan(x) | np.isnan(y)
                logger.debug(f"removing {np.sum(nan_m)} NaNs")

                corner.hist2d(
                    x[~nan_m],
                    y[~nan_m],
                    ax=axs[j][i],
                    plot_density=False,
                    color="k",
                    data_kwargs={"color": c},
                    contour_kwargs={"colors": c},
                    levels=1.0 - np.exp(-0.5 * np.array(level_sigma) ** 2)
                )

                axs[i][j].axis('off')

            if twinx[i] is None:
                twinx[i] = axs[i][i].twinx()

            twinx[i].hist(  # type: ignore
                data.loc[ids, color_keys[i]],
                color=c,
                density=True,
                histtype='step',
                bins=100
            )

            axs[i][0].set_ylabel(color_keys[i])
            axs[-1][i].set_xlabel(color_keys[i])

    return fig, axs


def agn_box_plot(
        base_name: str,
        database_name: str,
        wise_data: WISEDataDESYCluster,
        status: Status | None = None,
        colors: list[str] | None = None,
        level_sigma: list[float] | None = None,
):
    """
    Makes a color-color plot in the W1-W2 - W2-W3 plane and indicates the `AGN-Wedge`
    `(Hvding et al. (2022) <https://iopscience.iop.org/article/10.3847/1538-3881/ac5e33)>`_.

    :param base_name: base name for storage directories
    :type base_name: str
    :param database_name: name of the database
    :type database_name: str
    :param wise_data: instance of WISEDataDESYCluster
    :type wise_data: WISEDataDESYCluster
    :param status: statuses to include in the plot, defaults to ``["1", "2"]``
    :type status: list, optional
    :param colors: colors to use for the different statuses, defaults to ``["g", "purple"]``
    :type colors: list, optional
    :param level_sigma: sigma levels for the contours, defaults to ``[1, 5]``
    :type level_sigma: list, optional
    """

    logger.info("Making AGN-box color plot")

    if status is None:
        status = ["1", "2"]

    if colors is None:
        colors = ["g", "purple"]

    if level_sigma is None:
        level_sigma = [1, 5]

    parent_sample = wise_data.parent_sample
    wise_mag_keys = ["W1_mag", "W2_mag", "W3_mag"]

    if np.any([k not in parent_sample.default_keymap for k in wise_mag_keys]):
        logger.warning(f"One of {','.join(wise_mag_keys)} not in ParentSample's default keymap: \n"
                       f"{parent_sample.default_keymap}.\n"
                       f"Not making AGN-Box plot.")
        return

    agn_mask, W1W2, W2W3 = get_agn_mask_from_color(base_name, wise_data)
    parent_sample.df["W1-W2"] = W1W2
    parent_sample.df["W2-W3"] = W2W3

    n_agn = dict()

    for istatus in status:
        ids = DatabaseConnector(base_name=base_name, database_name=database_name).get_ids(istatus)
        status_agns = agn_mask.loc[ids]
        in_agns = np.sum(status_agns) / len(status_agns)
        logger.info(f"Status {istatus}: {in_agns*100:.2f}% in AGN box")
        n_agn[istatus] = in_agns

    fig, axs = color_plot(
        base_name,
        database_name,
        wise_data,
        status=status,
        colors=colors,
        color_keys=['W2-W3', 'W1-W2'],
        level_sigma=level_sigma,
        data=parent_sample.df
    )

    axs[0][0].set_xlim(-1, 7)
    axs[0][0].set_ylim(-1, 3)

    add_agn_box(axs[1][0])
    axs[1][0].legend()

    lines = [mpl.lines.Line2D([0], [0], color=c) for c in colors]
    labels = [f"{s} ({n_agn[s] * 100:.0f}% in AGN box)" for s in status]
    fig.legend(lines, labels, loc='lower left', fontsize="medium", bbox_to_anchor=(0.56, 0.56))

    fn = os.path.join(plots_dir("color_plots", base_name), "agn_box.pdf")
    logger.debug(f"saving under {fn}")
    fig.savefig(fn)

    plt.close()



"""
Module with functions to calculate the IR flux from the IR spectral flux densities.

* :func:`get_ban_nu` returns the frequency of a given band
* :func:`ir_flux_density_integral` integrates the flux density over the wavelength range of a given band
* :func: `calculate_ir_flux` calculates the IR flux for given indices
* :func:`get_ir_flux_index` get the calculated IR flux for given indices and cache them to disk
* :func:`get_ir_flux` get the IR flux for a given status
"""

import logging
import os
import json
import numpy as np
from tqdm import tqdm
from numpy import typing as npt
import pandas as pd
from astropy import units as u
from astropy import constants
from timewise.wise_data_base import WISEDataBase

from timewise_sup.environment import load_environment
from timewise_sup.mongo import DatabaseConnector, Index, Status, jd_offset
from timewise_sup.meta_analysis.baseline_subtraction import get_lightcurves


logger = logging.getLogger(__name__)


flux_key = "nuFnu_erg_per_s_per_sqcm"
flux_err_key = "nuFnu_err_erg_per_s_per_sqcm"
band_wavelengths = {
    # from Wright et al. (2010) ( 10.1088/0004-6256/140/6/1868 )
    "W1": 3.4 * 1e-6 * u.m,
    "W2": 4.6 * 1e-6 * u.m,
}


def get_band_nu(band: str | u.Quantity) -> u.Quantity:
    """Returns the frequency of a given band"""
    wl = band_wavelengths[band] if isinstance(band, str) else band
    return constants.c / wl


def flux_integral(
        spectral_flux_density: list[float],
        spectral_flux_density_unit: str,
        band: str | u.Quantity,
        out_unit: str = "erg s-1 cm-2"
) -> npt.NDArray:
    """
    Calculates the flux density value of a given spectral flux density. The flux density is calculated by multiplying
    the spectral flux density with the frequency of the given band.

    :param spectral_flux_density:
    :type spectral_flux_density: list[float]
    :param spectral_flux_density_unit:
    :type spectral_flux_density_unit: str
    :param band: either of "W1" or "W2" or the wavelength of the band
    :type band: str | astropy.units.Quantity
    :param out_unit:
    :type out_unit: str
    :return:
    """
    _flux = np.array(spectral_flux_density) * u.Unit(spectral_flux_density_unit)
    nu = get_band_nu(band)
    return np.array(u.Quantity(_flux * nu).to(out_unit).value)


def calculate_ir_flux(
        base_name: str,
        database_name: str,
        wise_data: WISEDataBase,
        index: Index,
        service: str = "tap",
        load_from_bigdata_dir: bool = False
) -> dict:
    """
    Calculates the IR flux for a given index. The IR flux is calculated by multiplying the flux density with the
    frequency of the given band.

    :param base_name: base name for storage directories
    :type base_name: str
    :param database_name: name of the database
    :type database_name: str
    :param wise_data: instance of WISEDataBase
    :type wise_data: WISEDataBase
    :param index: index of the object
    :type index: Index
    :param service: service to use for the TAP query, defaults to "tap", passed to :func:`get_lightcurves`
    :type service: str, optional
    :param load_from_bigdata_dir: load from the bigdata directory, defaults to False, passed to :func:`get_lightcurves`
    :type load_from_bigdata_dir: bool, optional
    :return: dictionary with the index as key and the IR fluxes as value
    :rtype: dict
    """
    indices = tuple(np.atleast_1d(index).astype(int))
    logger.info(f"calculating flux for {len(indices)} objects")

    lcs = get_lightcurves(base_name, database_name, wise_data, indices,
                          service=service, load_from_bigdata_dir=load_from_bigdata_dir)
    logger.debug(f"got {len(lcs)} lightcurves")
    lcs_with_fluxes = dict()

    for i, lc_dict in lcs.items():
        lc_in = pd.DataFrame.from_dict(lc_dict, orient="columns")
        lc_out = lc_in[["mean_mjd"]].copy()

        for b in ["W1", "W2"]:
            lc_out[f"{b}_{flux_key}"] = flux_integral(
                lc_in[f"{b}_diff_mean_flux_density"],
                spectral_flux_density_unit="mJy",
                band=b,
                out_unit="erg s-1 cm-2"
            )
            lc_out[f"{b}_{flux_err_key}"] = flux_integral(
                lc_in[f"{b}_diff_flux_density_rms"],
                spectral_flux_density_unit="mJy",
                band=b,
                out_unit="erg s-1 cm-2"
            )

        lcs_with_fluxes[str(i)] = lc_out.to_dict()

    return lcs_with_fluxes


def get_ir_flux(
        base_name: str,
        database_name: str,
        wise_data: WISEDataBase,
        status: Status,
        force_new: bool = False,
        service: str = "tap",
        load_from_bigdata_dir: bool = False
) -> dict:
    """
    Get the IR flux of the flares per status. If the cache file `flux_ir_lcs_status{status}.json` exists,
    it is loaded from there. If it does not exist, it is calculated by :func:`calculate_ir_flux` and stored there.

    :param base_name: base name for storage directories
    :type base_name: str
    :param database_name: name of the database
    :type database_name: str
    :param wise_data: instance of WISEDataBase
    :type wise_data: WISEDataBase
    :param status: status for which the IR flux should be calculated
    :type status: Status
    :param force_new: if True, the IR flux is calculated even if the cache file exists, defaults to False
    :type force_new: bool, optional
    :param service: service to use for the TAP query, defaults to "tap", passed to :func:`get_lightcurves`
    :type service: str, optional
    :param load_from_bigdata_dir: load from the bigdata directory, defaults to False, passed to :func:`get_lightcurves`
    :type load_from_bigdata_dir: bool, optional
    :return: dictionary with the index as key and the IR fluxes as value
    :rtype: dict
    """
    logger.info(f"getting fluxes for status {status}")
    tsup_data_dir = load_environment("TIMEWISE_SUP_DATA")
    fn = os.path.join(tsup_data_dir, base_name, f"flux_ir_lcs_status{status}.json")

    if (not os.path.isfile(fn)) or force_new:
        logger.debug(f"No file {fn}")
        logger.info("calculating fluxes")

        database_connector = DatabaseConnector(base_name=base_name, database_name=database_name)
        indices = database_connector.get_ids(status)
        lcs = calculate_ir_flux(
            base_name,
            database_name,
            wise_data,
            indices,
            service=service,
            load_from_bigdata_dir=load_from_bigdata_dir
        )

        logger.debug(f"writing to {fn}")
        with open(fn, "w") as f:
            json.dump(lcs, f)

    else:
        logger.debug(f"reading from {fn}")
        with open(fn, "r") as f:
            lcs = json.load(f)

    return lcs


def get_ir_flux_index(
        base_name: str,
        database_name: str,
        wise_data: WISEDataBase,
        index: Index,
        service: str = "tap",
        load_from_bigdata_dir: bool = False
) -> dict:
    """
    Get the IR flux of the flares by index. Checks the statuses of the flares and loads the respective files
    (or creates them) with :func:`get_ir_flux`.

    :param base_name: base name for storage directories
    :type base_name: str
    :param database_name: name of the database
    :type database_name: str
    :param wise_data: instance of WISEDataBase
    :type wise_data: WISEDataBase
    :param index: index of the object
    :type index: Index
    :param service: service to use for the TAP query, defaults to "tap", passed to :func:`get_lightcurves`
    :type service: str, optional
    :param load_from_bigdata_dir: load from the bigdata directory, defaults to False, passed to :func:`get_lightcurves`
    :type load_from_bigdata_dir: bool, optional
    :return: dictionary with the index as key and the IR fluxes as value
    :rtype: dict
    """
    _index = np.atleast_1d(index).astype(int)
    database_connector = DatabaseConnector(base_name=base_name, database_name=database_name)
    statuses = database_connector.get_status(list(_index))

    fluxes = dict()
    for status in statuses.status.unique():
        m = statuses.status == status
        status_ir_fluxes = get_ir_flux(base_name, database_name, wise_data, status,
                                       service=service, load_from_bigdata_dir=load_from_bigdata_dir)
        fluxes.update({str(i): status_ir_fluxes[str(i)] for i in _index[m]})

    return fluxes


def get_max_flux(
        base_name: str,
        database_name: str,
        wise_data: WISEDataBase,
        index: Index,
        service: str = "tap",
        load_from_bigdata_dir: bool = False
) -> pd.DataFrame:
    """
    Get the maximum flux of the flares by index. Gets the time of the maximum flux from the database and
    returns the flux at that time.

    :param base_name: base name for storage directories
    :type base_name: str
    :param database_name: name of the database
    :type database_name: str
    :param wise_data: instance of WISEDataBase
    :type wise_data: WISEDataBase
    :param index: index of the object
    :type index: Index
    :param service: service to use for the TAP query, defaults to "tap", passed to :func:`get_lightcurves`
    :type service: str, optional
    :param load_from_bigdata_dir: load from the bigdata directory, defaults to False, passed to :func:`get_lightcurves`
    :type load_from_bigdata_dir: bool, optional
    :return: dictionary with the index as key and the maximum flux as value
    :rtype: dict
    """

    logger.info(f"getting max flux for {len(np.atleast_1d(index))} objects of {base_name}")
    database_connector = DatabaseConnector(base_name=base_name, database_name=database_name)
    jd_of_max_flux = database_connector.get_peak_time_jd(index)
    mjd_of_max_flux = jd_of_max_flux + jd_offset
    fluxes = get_ir_flux_index(base_name, database_name, wise_data, index,
                               service=service, load_from_bigdata_dir=load_from_bigdata_dir)
    max_fluxes = {}
    for i in tqdm(index, desc="getting max fluxes"):
        lc = pd.DataFrame(fluxes[str(i)])
        i_max_fluxes = {}
        for b in ["W1", "W2"]:
            m = (lc.mean_mjd - mjd_of_max_flux.loc[str(i), f"max_mag_jd_{b}"]).abs() < 1e-9
            i_max_fluxes[f"{b}_max_flux"] = lc[m][f"{b}_{flux_key}"].values[0]
            i_max_fluxes[f"{b}_max_flux_err"] = lc[m][f"{b}_{flux_err_key}"].values[0]

        max_fluxes[str(i)] = i_max_fluxes

    return pd.DataFrame.from_dict(max_fluxes, orient="index")



"""
Calculates the IR luminosities of the lightcurves in the database.

* :func:`get_band_nu` returns the frequency of a given band
* :func:`nuFnu` calculates the flux density value of a given spectral flux density
* :func:`calculate_ir_luminosities` calculates the IR luminosities for a given index
* :func:`get_ir_luminosities` calculates the IR luminosities for a given status
"""

import logging
import os
import json

import numpy as np
import pandas as pd
from astropy import units as u
from astropy.cosmology import Planck18
from astropy.uncertainty import normal
from timewise.wise_data_base import WISEDataBase

from timewise_sup.environment import load_environment
from timewise_sup.mongo import DatabaseConnector, Index, Status, jd_offset
from timewise_sup.meta_analysis.flux import get_ir_flux_index, flux_key, flux_err_key


logger = logging.getLogger(__name__)

luminosity_key = "ir_luminosity_erg_per_s"
luminosity_err_key = "ir_luminosity_err_erg_per_s"
ref_time_key = "source_frame_time"


def get_ref_time(
        base_name: str,
        database_name: str,
        index: Index,
):
    connector = DatabaseConnector(base_name=base_name, database_name=database_name)
    peak_times = connector.get_peak_time_jd(index)
    mean_peak_time = peak_times.mean(axis=1)
    return mean_peak_time + jd_offset


def calculate_ir_luminosities(
        base_name: str,
        database_name: str,
        wise_data: WISEDataBase,
        index: Index,
        redshift: pd.Series,
        redshift_err: pd.Series,
) -> dict:
    """
    Calculates the IR luminosities for a given index. The IR luminosity is calculated by multiplying the flux density
    with the area of the sphere with the radius of the luminosity distance. The luminosity distance is calculated
    using the Planck18 cosmology. The area is calculated using the formula for the surface area of a sphere.

    :param base_name: base name for storage directories
    :type base_name: str
    :param database_name: name of the database
    :type database_name: str
    :param wise_data: instance of WISEDataBase
    :type wise_data: WISEDataBase
    :param index: index of the object
    :type index: Index
    :param redshift: redshift of the object
    :type redshift: pd.Series
    :param redshift_err: redshift error of the object
    :type redshift_err: pd.Series
    :return: dictionary with the index as key and the IR luminosities as value
    :rtype: dict
    """
    indices = tuple(np.atleast_1d(index).astype(int))
    logger.info(f"calculating luminosities for {len(indices)} objects")

    if len(indices) != len(redshift):
        raise ValueError("redshift and index must have the same length!")

    lcs = get_ir_flux_index(base_name, database_name, wise_data, indices)
    logger.debug(f"got {len(lcs)} lightcurves")

    ref_times = get_ref_time(base_name, database_name, indices)
    ref_times.index = ref_times.index.astype(str)
    lcs_with_luminosities = dict()

    # avoid SettingWithCopyWarning which falsely flags something like lc[ref_time_key] = ...
    with pd.option_context('mode.chained_assignment', None):

        for i, lc_dict in lcs.items():
            lc_in = pd.DataFrame.from_dict(lc_dict, orient="columns")
            lc = lc_in[["mean_mjd"]]

            iredshift = redshift[i]
            iredshift_err = redshift_err[i]

            one_plus_z = 1 + iredshift
            one_plus_z_err = iredshift_err if not np.isnan(iredshift_err) else 0
            one_by_one_plus_z = 1 / one_plus_z

            dl = Planck18.luminosity_distance(iredshift)
            area = 4 * np.pi * dl ** 2

            # estimate the uncertainty of the luminosity distance by sampling the redshift distribution
            # if the uncertainty of the redshift is given
            if not np.isnan(iredshift_err):
                iredshift_dist = normal(iredshift, std=iredshift_err, n_samples=10000)
                lum_dist_std = np.std(Planck18.luminosity_distance(iredshift_dist.distribution))
                area_unc = 8 * np.pi * dl * lum_dist_std
            else:
                area_unc = 0 * (u.Mpc ** 2)

            lc[ref_time_key] = (lc["mean_mjd"] - ref_times.loc[i]) * one_by_one_plus_z

            for b in ["W1", "W2"]:

                flux_val = u.Quantity(lc_in[f"{b}_{flux_key}"] * u.Unit("erg s-1 cm-2"))
                flux_valerr = u.Quantity(lc_in[f"{b}_{flux_err_key}"] * u.Unit("erg s-1 cm-2"))

                lum = u.Quantity(flux_val * area * one_plus_z).to("erg s-1").value
                lum_err = u.Quantity(
                    np.sqrt(
                        (flux_valerr * area * one_plus_z) ** 2
                        + (flux_val * area_unc * one_plus_z) ** 2
                        + (flux_val * area * one_plus_z_err) ** 2
                    )
                ).to("erg s-1").value

                lc[f"{b}_{luminosity_key}"] = lum
                lc[f"{b}_{luminosity_err_key}"] = lum_err

            lum_dict = {
                "lightcurve": lc.to_dict(),
                "z": iredshift,
                "z_err": iredshift_err,
                "ref_time_mjd": ref_times.loc[i]
            }

            lcs_with_luminosities[str(i)] = lum_dict

    logger.debug(f"calculated luminosities for {len(lcs_with_luminosities)} objects")
    return lcs_with_luminosities


def get_ir_luminosities(
        base_name: str,
        database_name: str,
        wise_data: WISEDataBase,
        status: Status,
        force_new: bool = False,
        redshifts: pd.DataFrame | None = None,
) -> dict:
    """
    Calculates the IR luminosities of the lightcurves in the database for a given status. If the cache file
    `lum_ir_lcs_status{status}.json` exists, it is loaded from there. If it does not exist, it is calculated by
    :func:`calculate_ir_luminosities` and stored there.

    :param base_name: base name for storage directories
    :type base_name: str
    :param database_name: name of the database
    :type database_name: str
    :param wise_data: instance of WISEDataBase
    :type wise_data: WISEDataBase
    :param status: status for which the IR luminosities should be calculated
    :type status: Status
    :param force_new: if True, the luminosities are calculated even if the cache file exists
    :type force_new: bool
    :param redshifts:
        redshift of the object, defaults to getting the redshift from the AMPEL catalog crossmatch,
        needs keys `z` and `z_unc`
    :type redshifts: pandas.DataFrame | None
    :return: dictionary with the index as key and the IR luminosities as value
    :rtype: dict
    """

    logger.info(f"getting luminosities for status {status}")
    tsup_data_dir = load_environment("TIMEWISE_SUP_DATA")
    fn = os.path.join(tsup_data_dir, base_name, f"lum_ir_lcs_status{status}.json")

    database_connector = DatabaseConnector(base_name=base_name, database_name=database_name)
    indices = database_connector.get_ids(status)

    if (not os.path.isfile(fn)) or force_new:
        logger.debug(f"No file {fn}")
        logger.info("calculating luminosities")

        if redshifts is None:
            redshifts = database_connector.get_redshift(indices)
            redshifts.rename(columns={"ampel_z": "z", "group_z_precision": "z_unc"}, inplace=True)

        lcs = calculate_ir_luminosities(
            base_name,
            database_name,
            wise_data,
            redshifts.index,
            redshifts["z"],
            redshifts["z_unc"]
        )

        missing_mask = ~pd.Index(indices).astype(str).isin(lcs.keys())
        if any(missing_mask):
            logger.warning(f"missing {missing_mask.sum()} lightcurves")
            missing_indices = np.array(indices)[missing_mask]
            for i in missing_indices:
                lcs[str(i)] = None

        logger.debug(f"writing to {fn}")
        with open(fn, "w") as f:
            json.dump(lcs, f)

    else:
        logger.debug(f"reading from {fn}")
        with open(fn, "r") as f:
            lcs = json.load(f)

        for i, lum_info in lcs.items():
            if lum_info is None:
                logger.warning(f"missing luminosity for {i}")

            elif (redshifts is not None) and (abs(1 - lum_info["z"] / redshifts.loc[i, "z"]) > 1e-7):
                raise ValueError(f"redshift mismatch for {i}: {lum_info['z']} != {redshifts.loc[i, 'z']}")

    return lcs


def get_ir_luminosities_index(
        base_name: str,
        database_name: str,
        wise_data: WISEDataBase,
        index: Index,
        force_new: bool = False,
        redshifts: pd.DataFrame | None = None,
) -> dict:
    logger.info(f"getting luminosities for {len(np.atleast_1d(index))} indices")
    database_connector = DatabaseConnector(base_name=base_name, database_name=database_name)
    statuses = database_connector.get_status(index)
    unique_statuses = statuses.status.unique()
    logger.debug(f"unique statuses: {unique_statuses}")

    lums_all = dict()

    for status in unique_statuses:
        lums = get_ir_luminosities(base_name, database_name, wise_data, status, force_new, redshifts)
        selected_lums = {k: v for k, v in lums.items() if k in statuses.index[statuses.status == status]}
        lums_all.update(selected_lums)

    logger.debug(f"got {len(lums_all)} lightcurves")
    return lums_all


# def get_peak_ir_luminosity_index(
#         base_name: str,
#         database_name: str,
#         index: Index,
# ) -> dict:
#     logger.info(f"getting peak IR luminosities for index {ind} ({base_name})")
#     lcs = get_ir_luminosities(base_name, database_name, status)
#     logger.debug(f"got {len(lcs)} lightcurves")
#
#     peak_lum = dict()
#     for i, lc_dict in lcs.items():
#         lc = pd.DataFrame.from_dict(lc_dict, orient="columns")
#         peak_lum[i] = dict()
#         for b in ["W1", "W2"]:
#             arg = np.argmax(lc[f"{b}_luminosity_erg_per_s"])
#             peak_lum[i][f"{b}_peak_luminosity_erg_per_s"] = lc[f"{b}_luminosity_erg_per_s"][arg]
#             peak_lum[i][f"{b}_peak_luminosity_err_erg_per_s"] = lc[f"{b}_luminosity_err_erg_per_s"][arg]
#             peak_lum[i][f"{b}_peak_luminosity_mjd"] = lc["mean_mjd"][arg]




"""
Implements the creation of :class:`AMPEL` job files

* :class:`ampel_job_file_filename()`: centralised function to get job file filenames
* :class:`keys_in_list()`: utility function to check of any key in a list is in a list of lists
* :class:`make_ampel_job_file()`: Creates the actual job file
"""

import json
import logging
import os.path
import yaml
import copy
from timewise.wise_data_base import WISEDataBase

from timewise_sup.environment import load_environment
from timewise_sup.plots import plots_dir as get_plots_dir
from timewise_sup.ampel_conf import (
    create_ampel_config_file,
    ampel_conf_filename,
    get_catalogue_match_conf,
    get_local_catalogue_match_conf,
    get_filter_config
)


logger = logging.getLogger(__name__)


def ampel_job_file_filename(
        base_name,
        chunk,
        t2units,
        precut_filter,
):
    """Get the filename for a job file with the given arguments. See :class:`make_ampel_job_file()` for arguments."""
    units_str = "run"
    if keys_in_list(["bayesian", "dust"], t2units):
        units_str += "_bayeblocks"
    if keys_in_list(["redshift", "catalogue"], t2units):
        units_str += "_xmatch"
    if keys_in_list(["local_catalogue"], t2units):
        units_str += "_localxmatch"
    if keys_in_list(["redshift"], t2units):
        units_str += "_zdigest"

    tsup_data = load_environment("TIMEWISE_SUP_DATA")
    fn = os.path.join(
        tsup_data,
        base_name,
        "ampel_job_files",
        f"{chunk}__{units_str}__filter{precut_filter}.yml"
    )
    return fn


def keys_in_list(keys, arr):
    """Checks if any of the keys is in any of the lists"""
    return any([any([k in element for element in arr]) for k in keys])


def make_ampel_job_file(
        base_name: str,
        database_name: str,
        wise_data: WISEDataBase,
        chunk: int | None,
        t2units: list[str],
        precut_filter: bool | None,
        load_from_bigdata_dir: bool = True,
        service: str = "tap",
        concurrent: bool = False,
        split_units: bool = False
):
    """
    Creates the :class:`AMPEL` job file running the specified units.

    :param base_name: base name of the analysis
    :type base_name: str
    :param database_name: Name of the database in the MongoDB
    :type database_name: str
    :param wise_data: The wise data to be analysed
    :type wise_data: timewise.wise_data_base.WISEDataBase
    :param chunk: The chunk of the wise data
    :type chunk: int
    :param t2units:
        The :class:`AMPEL` tier 2 units to run
        (see the `Ampel-hu-astro repo <https://github.com/AmpelAstro/Ampel-HU-astro>`_).

        * :code:`"bayesian"`, :code:`"dust"`:
            Runs :class:`T2BayesianBlocks` to find excess regions and :class:`T2DustEchoEval` to characterise the result.
        * :code:`"catalogue"`: :class:`T2CatalogMatch`
        * :code:`"redshift"`: :class:`T2CatalogMatch` and :class:`T2DigestRedshifts`
        * :code:`"catalogue_local"`: :class:`T2CatalogMatchLocal`

    :type t2units: list[str]
    :param load_from_bigdata_dir: If :code:`True` loads from the wise data's big data directory
    :type load_from_bigdata_dir: bool
    :param service: The :class:`timewise` service used to download the data, either :code:`"tap"` or :class:`"gator"`
    :type service: str
    :param precut_filter: Run :class:`T2DustEchoEval` only for lightcurves with
        significant variation compared to the median :math:`\chi^2/\mathrm{ndoF} > 1`
    :type precut_filter: bool
    :param concurrent: Optimised set-up for running :class:`AMPEL` concurrently
    :type concurrent: bool
    :param split_units: Split units in different tasks
    :type split_units: bool
    :return: Filename of the job file
    :rtype: str
    """
    logger.info(f"making ampel job file for chunk {chunk} of {base_name}")

    # create ampel config file
    AMPEL_CONF = ampel_conf_filename()
    if not os.path.isfile(AMPEL_CONF):
        create_ampel_config_file()

    # ------------------- assemble configurations for T2 units ------------------- #
    # --------------------------------  START  ----------------------------------- #
    #

    # T2CatalogueMatch
    match_dist = 5.
    t2_catalogue_match_config = get_catalogue_match_conf(match_dist)

    # T2LocalCatalogueMatch
    t2_local_cataklogue_match_config = get_local_catalogue_match_conf(match_dist)

    # T2DigestRedshifts
    t2_digest_redshifts_config = {
        "max_redshift_category": 7,
        "t2_dependency": [
            {"unit": "T2CatalogMatch", "config": t2_catalogue_match_config, "link_override": {"filter": "PPSFilter"}}
        ]
    }

    # T2BayesianBlocks
    plots_dir = get_plots_dir("ampel_plots", base_name=base_name)
    bayesian_blocks_db_format = {
        "channel": "wise",
        "BayePlots": False,
        "BayePlotsDir": os.path.join(plots_dir, "T2BayesianBlocks"),
        "DustEchoDir": os.path.join(plots_dir, "T2DustEchoEval")
    }
    t2_bayesian_blocks_config = {
        "debug": logger.getEffectiveLevel() <= logging.DEBUG,
        "debug_dir": bayesian_blocks_db_format["BayePlotsDir"],
        "rej_sigma": 5,
        "plot": bayesian_blocks_db_format["BayePlots"],
        "data_type": "wise",
        "filters": ["Wise_W1", "Wise_W2"],
        "flux": True,
        "Npoints": True,
        "plot_props": {
            "file_name": {"format_str": "%s_%s.svg", "arg_keys": ["stock", "band"]},
            "title": {"format_str": "%s", "arg_keys": ["stock"]},
            "width": 18,
            "height": 23,
            "fig_include_title": False,
            "disk_save": bayesian_blocks_db_format["BayePlotsDir"],
        },
    }

    # T2DustEchoEval
    t2_dust_echo_eval_config = {
        "flux": True,
        "filters": ["Wise_W1", "Wise_W2"],
        "directory": bayesian_blocks_db_format["DustEchoDir"],
        "t2_dependency": [
            {
                "unit": "T2BayesianBlocks",
                "config": t2_bayesian_blocks_config,
            },
        ],
    }

    #
    # ------------------- assemble configurations for T2 units ------------------- #
    # --------------------------------   END   ----------------------------------- #

    # --------------------- assemble directives for T2 units --------------------- #
    # --------------------------------  START  ----------------------------------- #

    # find units to use
    units = [e.lower() for e in t2units]
    directives = list()
    unit_ids = list()
    channels = list()

    # ------- bayesian blocks ------- #

    # check if any of the units specify the T2BayesianBlocks + T2DustEchoEval configuration
    if keys_in_list(["bayesian", "dust"], units):
        filter_config = get_filter_config(logical_connection="AND", red_chi2_threshold=1 if precut_filter else None)
        t2_bayesian_blocks_directives = {
            "channel": bayesian_blocks_db_format["channel"],
            "filter": filter_config,
            "ingest": {
                "mux": {
                    "unit": "ZiMongoMuxer",
                    'config': {"db_complete": False},
                    "combine": [{
                        "unit": "ZiT1Combiner",
                        "state_t2": [
                            {"unit": "T2BayesianBlocks", "config": t2_bayesian_blocks_config},
                            {"unit": "T2DustEchoEval", "config": t2_dust_echo_eval_config}
                        ]
                    }],
                }
            }
        }

        directives.append(t2_bayesian_blocks_directives)
        channels.append(bayesian_blocks_db_format["channel"])
        unit_ids.extend(["T2BayesianBlocks", "T2DustEchoEval"])

    # ------- Catalogue match and Redshift digest  ------- #

    if keys_in_list(["redshift", "catalogue", "catalogue_local"], units):
        catalogue_redshift_channel = "redshift_catalog_match"
        channels.append(catalogue_redshift_channel)

        point_t2 = [{
            "unit": "T2CatalogMatch",
            "config": t2_catalogue_match_config,
            "ingest": {"filter": "PPSFilter", "sort": "jd", "select": "first"},
        }]

        state_t2 = []

        if keys_in_list(["redshift"], units):
            state_t2.append({'unit': 'T2DigestRedshifts', 'config': t2_digest_redshifts_config})

        if keys_in_list(["local"], units):
            state_t2.append({"unit": "T2CatalogMatchLocal", "config": t2_local_cataklogue_match_config})

        combine = [{
            "unit": "ZiT1Combiner",
            "point_t2": point_t2,
            "state_t2": state_t2
        }]

        t2_catalogue_match_directives = {
            "channel": catalogue_redshift_channel,
            "filter": get_filter_config(logical_connection="OR"),
            "ingest": {
                "mux": {
                    "unit": "ZiMongoMuxer",
                    "config": {"db_complete": False},
                    "combine": combine,
                }
            }
        }

        directives.append(t2_catalogue_match_directives)

        # check which exact units to run

        if keys_in_list(["redshift", "catalogue"], units):
            unit_ids.append("T2CatalogMatch")

        if keys_in_list(["catalogue_local"], units):
            unit_ids.append("T2LocalCatalogMatch")

        if keys_in_list(["redshift"], units):
            unit_ids.append("T2DigestRedshifts")

    logger.debug(f"unit ids: {unit_ids}")

    #
    # --------------------- assemble directives for T2 units --------------------- #
    # --------------------------------   END   ----------------------------------- #

    # --------------------------- make AMPEL job file ---------------------------- #
    # --------------------------------  START  ----------------------------------- #
    #

    if load_from_bigdata_dir:
        alert_archive = wise_data._data_product_filename(service=service, chunk_number=chunk, use_bigdata_dir=True)
    else:
        alert_archive = wise_data._data_product_filename(service=service, chunk_number=chunk)

    channel_info = [{
        "name": c,
        "version": 0,
        "access": ["ZTF_PUB"],
        "policy": []
    } for c in channels]

    ampel_job_dictionary = {
        "name": 'timewise_sup',
        "mongo":
            {
                "prefix": database_name,
                "reset": False,
            },
        "channel": channel_info,
        "task": [
            {
                "title": 't0',
                "unit": "AlertConsumer",
                "multiplier": 1,
                "config": {
                    "log_profile": "prod",
                    "compiler_opts": "ZiCompilerOptions",
                    "shaper": "ZiDataPointShaper",
                    "supplier": {
                        "unit": "NeoWisePhotometryAlertSupplier",
                        "config": {
                            "dpid": "hash",
                            "loader": {
                                "unit": "WiseFileAlertLoader",
                                "config": {
                                    "file": alert_archive,
                                }
                            }
                        }
                    },
                    "directives": directives,
                    "iter_max": 1000000
                }
            }
        ]
    }

    t2_task_template = {
        'title': "t2",
        'unit': 'T2Worker',
        'multiplier': 1,
        'config': {
            "log_profile": "default",
        }
    }

    if concurrent:
        t2_task_template["config"] = {
            "doc_limit": 1000000,
            "pre_check": False,
            "run_dependent_t2s": False,
            "wait_for_durable_write": False,
            "updates_buffer_size": 500,
            "garbage_collect": False,
            "max_try": 10,
            "log_profile": "prod",
            "code_match": [-1, -6, -2000, -2001, -2004, -2006, -2008],
            "pending_codes": [-2001, -2006, -2004, -2008, -5, -1, -2000, -7],
        }

    if split_units:
        for unit_id in unit_ids:
            task = copy.deepcopy(dict(t2_task_template))  # type: dict
            task["config"]["unit_ids"] = [unit_id]
            logger.debug(f"adding task {json.dumps(task, indent=4)}")
            ampel_job_dictionary["task"].append(task)  # type: ignore

    else:
        logger.debug(f"adding task {json.dumps(t2_task_template, indent=4)}")
        ampel_job_dictionary["task"].append(t2_task_template)  # type: ignore

    filename = ampel_job_file_filename(
        base_name=base_name,
        chunk=chunk,
        t2units=t2units,
        precut_filter=precut_filter
    )

    logger.info(f"writing ampel job file to {filename}")
    d = os.path.dirname(filename)
    if not os.path.isdir(d):
        os.makedirs(d)
    with open(filename, "w") as f:
        yaml.dump(ampel_job_dictionary, f)

    return filename

    #
    # --------------------------- make AMPEL job file ---------------------------- #
    # --------------------------------   END   ----------------------------------- #



import os.path
import argparse
import logging
import json

from timewise_sup.environment import load_environment


tsup_data_dir = load_environment("TIMEWISE_SUP_DATA")
logger = logging.getLogger(__name__)


bandcolors = {
    "W1": "purple",
    "W2": "plum",
    "ZTF_g": "forestgreen",
    "ZTF_r": "crimson",
    "ZTF_i": "khaki"
}


status_linestyle = {
    "1": "-",
    "2": "--",
    "2_maybe_interesting": ":",
    "3": "-.",
    "4": "-.",
    "No_further_investigation": "-.",
}


types_colors = {
    "agn": "mediumpurple",
    "tde": "darkviolet",
    "failed": "tomato",
    "sn_other": "darkturquoise",
    "snia": "deepskyblue",
    "star": "beige",
    "unclass": "lightgrey"
}


type_labels = {
    "agn": "AGN",
    "tde": "TDE",
    "sn_other": "SN ($\\ne$ Ia)",
    "snia": "SN Ia",
    "failed": "failed fit",
    "star": "Star",
    "unclass": "unclassified"
}


def plots_dir(module: str, base_name: str) -> str:
    d = os.path.join(tsup_data_dir, "plots", module, base_name)
    if not os.path.isdir(d):
        os.makedirs(d)
    return d



"""
This module contains functions for identifying AGN in WISE data.

* :func:`get_agn_mask_from_color` uses the WISE color-color diagram from Hviding et al. (2022) to identify AGN
* :func:`add_agn_box` adds the AGN selection box from Hviding et al. (2022) to a plot
* :func:`get_agn_mask_from_variability` uses the variability in the IR lightcurves to identify AGN
"""

import matplotlib as mpl
import logging
import pandas as pd
from timewise.wise_data_base import WISEDataBase

from timewise_sup.mongo import DatabaseConnector


logger = logging.getLogger(__name__)


# taken from Hviding et al. (2022)
# https://iopscience.iop.org/article/10.3847/1538-3881/ac5e33
agn_box: dict[str, list] = {
    "W2-W3": [1.734, 3.916],
    "W1-W2 / W2-W3 parameters": [
        [0.0771, 0.319],
        [0.261, -0.260],
    ]
}


def get_agn_mask_from_color(
        base_name: str,
        wise_data: WISEDataBase
) -> tuple[pd.DataFrame, pd.Series, pd.Series]:
    """
    Get the AGN mask based on the WISE color-color diagram from Hviding et al. (2022)

    :param base_name: the base name of the WISE data
    :type base_name: str
    :param wise_data: the WISE data
    :type wise_data: WISEDataBase
    :return: the AGN mask, the W1-W2 color, and the W2-W3 color
    :rtype: tuple[pd.DataFrame, pd.Series, pd.Series]
    """
    logger.info(f"getting agn mask for {base_name}")
    parent_sample = wise_data.parent_sample

    wise_mag_keys = ["W1_mag", "W2_mag", "W3_mag"]
    mag_keys = {b: parent_sample.default_keymap[b] for b in wise_mag_keys}
    logger.debug(f"WISE filter mag keys are {mag_keys}")
    parent_sample.df["W1-W2"] = parent_sample.df[mag_keys["W1_mag"]] - parent_sample.df[mag_keys["W2_mag"]]
    parent_sample.df["W2-W3"] = parent_sample.df[mag_keys["W2_mag"]] - parent_sample.df[mag_keys["W3_mag"]]

    W1W2 = parent_sample.df["W1-W2"]
    W2W3 = parent_sample.df["W2-W3"]
    agn_mask = (
            (W2W3 > agn_box["W2-W3"][0]) &
            (W2W3 < agn_box["W2-W3"][1]) &
            (W1W2 > agn_box["W1-W2 / W2-W3 parameters"][0][0] * W2W3 + agn_box["W1-W2 / W2-W3 parameters"][0][1]) &
            (W1W2 > agn_box["W1-W2 / W2-W3 parameters"][1][0] * W2W3 + agn_box["W1-W2 / W2-W3 parameters"][1][1])
    )

    return agn_mask, W1W2, W2W3


def add_agn_box(
        ax: mpl.axes.Axes,
):
    """
    Add the AGN selection box from Hviding et al. (2022) to a plot

    :param ax: the axis to add the box to
    :type ax: mpl.axes.Axes
    """
    logger.info("adding AGN wedge to plot")
    agn_box_style = {
        "lw": 2,
        "ls": ":",
        "color": "k"
    }

    # store axis limits
    xlim = ax.get_xlim()
    ylim = ax.get_ylim()

    # calculate the point where the agn wedge lines meet
    x = [
        agn_box["W2-W3"][0],
        agn_box["W2-W3"][0],
        -(agn_box["W1-W2 / W2-W3 parameters"][0][1] - agn_box["W1-W2 / W2-W3 parameters"][1][1]) /
        (agn_box["W1-W2 / W2-W3 parameters"][0][0] - agn_box["W1-W2 / W2-W3 parameters"][1][0]),
        agn_box["W2-W3"][1],
        agn_box["W2-W3"][1],
    ]
    y = [
        1e3,
        agn_box["W1-W2 / W2-W3 parameters"][0][0] * x[1] + agn_box["W1-W2 / W2-W3 parameters"][0][1],
        agn_box["W1-W2 / W2-W3 parameters"][0][0] * x[2] + agn_box["W1-W2 / W2-W3 parameters"][0][1],
        agn_box["W1-W2 / W2-W3 parameters"][1][0] * x[3] + agn_box["W1-W2 / W2-W3 parameters"][1][1],
        1e3
    ]
    logger.debug(f"x: {x}")
    logger.debug(f"y: {y}")

    ax.plot(x, y, label="Hviding et al. (2022)", **agn_box_style)  # type: ignore

    # reset axis limits
    ax.set_xlim(xlim)
    ax.set_ylim(ylim)


def get_agn_mask_from_variability(
        base_name: str,
        database_name: str,
        wise_data: WISEDataBase,
) -> pd.DataFrame:
    """
    Get the AGN mask based on the variability in the IR lightcurves.

    :param base_name: the base name of the WISE data
    :type base_name: str
    :param database_name: the name of the database to get the variability stocks from
    :type database_name: str
    :param wise_data: the WISE data
    :type wise_data: WISEDataBase
    :return: the AGN mask
    :rtype: pd.DataFrame
    """
    logger.info(f"getting AGN mask based on variability")
    stocks = DatabaseConnector(base_name=base_name, database_name=database_name).get_agn_variability_stocks()
    logger.debug(f"got {len(stocks)}, making mask")
    mask = pd.DataFrame(
        {"variability_agn_mask": wise_data.parent_sample.df.index.isin(stocks)},
        index=wise_data.parent_sample.df.index
    )
    return mask



"""
This module implements a function that creates an :class:`AMPEL` job file and runs it

* :class:`bayesian_blocks()`: Creates and runs an :class:`AMPEL` job file
"""

import logging
import subprocess
from timewise.wise_data_base import WISEDataBase

from timewise_sup.ampel_conf import ampel_conf_filename
from timewise_sup.analyse_lightcurves.create_job_file_yaml import make_ampel_job_file

logger = logging.getLogger("timewise_sup.analyze_lightcurves.bayesian_blocks")


def bayesian_blocks(
    base_name: str,
    database_name: str,
    wise_data: WISEDataBase,
    chunk: int | None,
    t2units: list[str],
    load_from_bigdata_dir: bool = True,
    service: str = "tap",
    precut_filter: bool | None = False,
):
    """
    1. Uses :class:`make_ampel_job_file() to create an :class:`AMPEL` job file.
    2. Assumes that :class:`create_ampel_job_file()` has been run.
    3. Executes :code:`ampel job -config <config_file> -schema <job_file>`

    See :class:`timewise_sup.analyse_lightcurves.create_job_file_yaml.make_ampel_job_file()` for arguments.
    """

    fn = make_ampel_job_file(
        base_name=base_name,
        wise_data=wise_data,
        database_name=database_name,
        chunk=chunk,
        t2units=t2units,
        precut_filter=precut_filter,
        concurrent=False,
        split_units=False,
        load_from_bigdata_dir=load_from_bigdata_dir,
        service=service
    )

    logger.info("running ampel")
    with subprocess.Popen(["ampel", "job", "-config", ampel_conf_filename(), "-schema", fn]) as p:
        out, err = p.communicate()
        if out is not None:
            logger.info(out.decode())
        if err is not None:
            logger.error(err.decode())



"""
Functions for getting catalog match masks.

* :func:`get_catalog_match_mask` gets a mask for the catalog matches of a given base name.
"""

import logging
import pandas as pd

from timewise_sup.mongo import DatabaseConnector, Index


logger = logging.getLogger(__name__)


def get_catalog_match_mask(
        base_name: str,
        database_name: str,
        index: Index
) -> pd.DataFrame:
    """
    Get the catalog match mask for a given base name. True means that the object has a match in any of the catalogs.

    :param base_name: base name of the WISE data
    :type base_name: str
    :param database_name: name of the database
    :type database_name: str
    :param index: the index of the objects to get the catalog match mask for
    :type index: Index
    :return: the catalog match mask
    :rtype: pd.DataFrame
    """
    logger.info("getting catalog match mask")
    matches = DatabaseConnector(base_name=base_name, database_name=database_name).get_catalog_matches(index)
    has_match_dict = {i: {c: match is not None for c, match in imatches.items()} for i, imatches in matches.items()}
    has_match_masks = pd.DataFrame.from_dict(has_match_dict, orient="index").fillna(False)
    return has_match_masks



"""
Make a plot of the rise and fade times for all flares in the sample.

* :func:`make_rise_fade_time_plot` makes the plot
"""

import logging
import os

import pandas as pd
import seaborn as sns
import matplotlib.pyplot as plt
import numpy as np

from timewise_sup.plots import plots_dir, bandcolors, status_linestyle
from timewise_sup.mongo import DatabaseConnector


logger = logging.getLogger(__name__)


def make_rise_fade_time_plot(
        base_name: str,
        database_name: str
):
    """
    Make a plot of the rise and fade times for all flares in the sample.

    :param base_name: the base name of the sample
    :type base_name: str
    :param database_name: the name of the database
    :type database_name: str
    """
    logger.info("making rise/fade time plot")
    database_connector = DatabaseConnector(base_name=base_name, database_name=database_name)
    ids = database_connector.get_ids(("1", "2"))

    status = database_connector.get_status(tuple(ids))
    rise_times = database_connector.get_rise_time(ids)
    fade_times = database_connector.get_fade_time(ids)
    data = pd.concat([status, np.log10(rise_times), np.log10(fade_times)], axis=1)

    fig, axs = plt.subplots(
        nrows=1, ncols=2, sharex="all", sharey="all",
        gridspec_kw={"hspace": 0, "wspace": 0},
        figsize=(3*2, 5)
    )
    for b, ax in zip(["W1", "W2"], axs):
        # logger.debug(f"band {b}, status {i_status}, n={len(data)}")
        for i_status in ["1", "2"]:
            d = data[(data["status"] == i_status)]
            logger.debug(f"band {b}, status {i_status}, n={len(d)}")
            sns.kdeplot(
                data=d,
                x=f"e_rise_{b}",
                y=f"e_fade_{b}",
                ax=ax,
                color=bandcolors[b],
                alpha=1,
                linestyles=status_linestyle[i_status],
                levels=3,
                fill=False,
                hatches=["\\\\", "///", "xxx"],
                label=i_status,
                zorder=2
            )

        if len(data) < 100:
            sns.scatterplot(
                data=data,
                x=f"e_rise_{b}",
                y=f"e_fade_{b}",
                ax=ax,
                style="status",
                color=bandcolors[b],
                zorder=1
            )

        ax.set_aspect("equal", adjustable="box")
        ax.grid(ls=":", zorder=10, alpha=0.5)
        ax.set_xlabel("")
        ax.set_ylabel("")
        ax.legend()

    fig.supxlabel("rise time [log$_{10}$(days)]")
    fig.supylabel("fade time [log$_{10}$(days)]")
    fig.tight_layout()
    fn = os.path.join(plots_dir("rise_fade_times", base_name), "rise_fade_times.pdf")
    logger.info(f"saving under {fn}")
    fig.savefig(fn)
    plt.close()



import os
import logging
from pathlib import Path

from timewise.general import data_dir


logger = logging.getLogger(__name__)

environment_variable_defaults = {
    "TIMEWISE_SUP_DATA": str(Path(data_dir) / "subtraction_pipeline"),
    "TIMEWISE_SUP_MONGODB_PORT": "27017",
    "ZTF_FPH": None,
    "ZTF_NUCLEAR_SAMPLE_DB_PATH": None
}


def load_environment(key):
    try:
        return os.environ[key]
    except KeyError:
        default_val = environment_variable_defaults[key]
        logger.debug(f"environment variable {key} not specified. Using default {default_val}")
        return default_val



"""
Module with functions to calculate the IR fluence from the IR fluxes.

* :func:`ir_flux_integral` integrates the flux over the duration of the flare
* :func: `calculate_ir_fluence` calculates the IR fluence for given indices
* :func:`get_ir_fluence` get the calculated IR fluence for given status and cache them to disk
"""

import logging
import os
import json
import numpy as np
import pandas as pd
from timewise.wise_data_base import WISEDataBase

from timewise_sup.environment import load_environment
from timewise_sup.mongo import DatabaseConnector, Status
from timewise_sup.meta_analysis.flux import get_ir_flux, flux_key, flux_err_key
from timewise_sup.meta_analysis.integrate_time import time_integral, get_flare_time


logger = logging.getLogger(__name__)


fluence_key = "ir_fluence_erg_per_sqcm"
fluence_err_key = "ir_fluence_err_erg_per_sqcm"


def calculate_ir_fluence(
        base_name: str,
        database_name: str,
        wise_data: WISEDataBase,
        status: Status,
        service: str = "tap",
        load_from_bigdata_dir: bool = False
) -> dict:
    """
    Calculate the IR fluence for a given status. The IR fluence is calculated by integrating the IR flux over the
    duration of the flare.

    :param base_name: base name for storage directories
    :type base_name: str
    :param database_name: name of the database
    :type database_name: str
    :param wise_data: instance of WISEDataBase
    :type wise_data: WISEDataBase
    :param status: status to calculate the IR fluence for
    :type status: Status
    :param service: service to use for the TAP query, defaults to "tap", passed to :func:`get_lightcurves`
    :type service: str, optional
    :param load_from_bigdata_dir: load from the bigdata directory, defaults to False, passed to :func:`get_lightcurves`
    :type load_from_bigdata_dir: bool, optional
    :return:
    """
    database_connector = DatabaseConnector(base_name=base_name, database_name=database_name)
    ids = database_connector.get_ids(status)
    indices = np.atleast_1d(ids)
    logger.info(f"calculating fluence for {len(indices)} objects")
    lcs = get_ir_flux(base_name, database_name, wise_data, status,
                      service=service, load_from_bigdata_dir=load_from_bigdata_dir)
    flare_time = get_flare_time(base_name, database_name, indices)
    fluence = dict()

    for k, lc in lcs.items():
        i_fluence = time_integral(
            lc=pd.DataFrame.from_dict(lc, orient="columns"),
            key=flux_key,
            err_key=flux_err_key,
            t_start=flare_time.loc[str(k), "start_time"],
            t_end=flare_time.loc[str(k), "end_time"],
        )
        for b in ["W1", "W2"]:
            i_fluence[f"{b}_{fluence_key}"] = i_fluence.pop(f"{b}_{flux_key}_integrated")
            i_fluence[f"{b}_{fluence_err_key}"] = i_fluence.pop(f"{b}_{flux_key}_integrated_err")

        i_fluence["ir_fluence_is_lower_limit"] = bool(flare_time.loc[k, "lower_limit"])
        fluence[k] = i_fluence

    return fluence


def get_ir_fluence(
        base_name: str,
        database_name: str,
        wise_data: WISEDataBase,
        status: Status,
        service: str = "tap",
        load_from_bigdata_dir: bool = False,
        force_new: bool = False
) -> dict:
    """
    Get the IR fluence for a given status. The IR fluence is calculated by integrating :func:`get_ir_flux` if it
    has not been calculated before. The results are cached to disk.

    :param base_name: base name for storage directories
    :type base_name: str
    :param database_name: name of the database
    :type database_name: str
    :param wise_data: instance of WISEDataBase
    :type wise_data: WISEDataBase
    :param status: status to calculate the IR fluence for
    :type status: Status
    :param service: service to use for the TAP query, defaults to "tap", passed to :func:`get_lightcurves`
    :type service: str, optional
    :param load_from_bigdata_dir: load from the bigdata directory, defaults to False, passed to :func:`get_lightcurves`
    :type load_from_bigdata_dir: bool, optional
    :param force_new: calculate fluence again, even if cache exists, defaults to False
    :type force_new: bool, optional
    :return: dictionary with the IR fluence for each index
    :rtype: dict
    """

    logger.info(f"getting fluxes for status {status}")
    tsup_data_dir = load_environment("TIMEWISE_SUP_DATA")
    fn = os.path.join(tsup_data_dir, base_name, f"fluence_ir_status{status}.json")
    logger.info(f"getting IR fluence for status {status}")

    if os.path.exists(fn) and not force_new:
        logger.info(f"loading IR fluence from {fn}")
        with open(fn, "r") as f:
            fluence = json.load(f)

    else:
        logger.debug(f"No file {fn}")
        logger.info(f"calculating IR fluence")
        fluence = calculate_ir_fluence(
            base_name=base_name,
            database_name=database_name,
            wise_data=wise_data,
            status=status,
            service=service,
            load_from_bigdata_dir=load_from_bigdata_dir
        )
        logger.info(f"saving IR fluence to {fn}")
        with open(fn, "w") as f:
            json.dump(fluence, f)

    return fluence
